#
# This file is part of PCAP BGP Parser (pbgpp)
#
# Copyright 2016-2017 DE-CIX Management GmbH
# Author: Tobias Hannaske <tobias.hannaske@de-cix.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from binascii import hexlify

import datetime

from pbgpp.PCAP.Exceptions import PCAPInformationError


class PCAPInformation:
    def __init__(self, ts, mac, ip, ports):
        if not isinstance(mac, PCAPLayer2Information):
            raise PCAPInformationError("parameter 'mac' must be instance of PCAPLayer2Information")

        if not isinstance(ip, PCAPLayer3Information):
            raise PCAPInformationError("parameter 'ip' must be instance of PCAPLayer3Information")

        if not isinstance(ports, PCAPLayer4Information):
            raise PCAPInformationError("parameter 'ports' must be instance of PCAPLayer4Information")

        # Assign variables
        self.ts = ts # (Timestamp)
        self.mac = mac # (MAC address set)
        self.ip = ip # (IP address set)
        self.ports = ports # (Port set)

    def get_timestamp(self):
        return self.ts

    def get_timestmap_utc(self):
        return datetime.datetime.utcfromtimestamp(self.ts[0]).strftime('%Y-%m-%d %H:%M:%S') + "." + str(self.ts[1])

    def get_mac(self):
        return self.mac

    def get_ip(self):
        return self.ip

    def get_ports(self):
        return self.ports

    def get_source_mac(self):
        return self.mac.source

    def get_source_ip(self):
        return self.ip.source

    def get_source_port(self):
        return self.ports.source

    def get_destination_mac(self):
        return self.mac.destination

    def get_destination_ip(self):
        return self.ip.destination

    def get_destination_port(self):
        return self.ports.destination


class PCAPLayer2Information:
    def __init__(self, source, destination):
        # Store source and destination MAC address
        self.source = source
        self.destination = destination

    def get_source_string(self, separated=False):
        if self.source is None:
            output = "000000000000"
        else:
            output = str(hexlify(self.source).decode("utf-8"))

        if separated:
            return output[0:2] + ":" + output[2:4] + ":" + output[4:6] + ":" + output[6:8] + ":" + output[8:10] + ":" + output[10:12]
        else:
            return output

    def get_destination_string(self, separated=False):
        if self.destination is None:
            output = "000000000000"
        else:
            output = str(hexlify(self.destination).decode("utf-8"))

        if separated:
            return output[0:2] + ":" + output[2:4] + ":" + output[4:6] + ":" + output[6:8] + ":" + output[8:10] + ":" + output[10:12]
        else:
            return output

    def __str__(self):
        return "<PCAPLayer2Information source={0} destination={1}>".format(self.get_source_string(), self.get_destination_string())


class PCAPLayer3Information:
    def __init__(self, source, destination):
        # Store source and destination IP address
        self.source = source
        self.destination = destination

    def get_source_string(self):
        return str(self.source[0]) + "." + str(self.source[1]) + "." + str(self.source[2]) + "." + str(self.source[3])

    def get_destination_string(self):
        return str(self.destination[0]) + "." + str(self.destination[1]) + "." + str(self.destination[2]) + "." + str(self.destination[3])

    def __str__(self):
        return "<PCAPLayer3Information source={0} destination={1}>".format(self.get_source_string(), self.get_destination_string())


class PCAPLayer4Information:
    def __init__(self, source, destination):
        # Store source and destination port
        self.source = source
        self.destination = destination

    def get_source_string(self):
        return str(self.source)

    def get_destination_string(self):
        return str(self.destination)

    def __str__(self):
        return "<PCAPLayer4Information source={0} destination={1}>".format(self.get_source_string(), self.get_destination_string())



#
# This file is part of PCAP BGP Parser (pbgpp)
#
# Copyright 2016-2017 DE-CIX Management GmbH
# Author: Tobias Hannaske <tobias.hannaske@de-cix.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from pbgpp.BGP.Open.OptionalParameter import BGPOptionalParameter
from pbgpp.BGP.Statics import BGPStatics
from pbgpp.BGP.Translation import BGPTranslation


class BGPOptionalParameterReserved(BGPOptionalParameter):

    def __init__(self, payload):
        BGPOptionalParameter.__init__(self, payload)
        self.type = BGPStatics.OPEN_RESERVED
        self.__parse()

    def __parse(self):
        self.parsed = True

    def json(self):
        return {
            "type": self.type,
            "type_string": BGPTranslation.open_parameter(self.type)
        }



#
# This file is part of PCAP BGP Parser (pbgpp)
#
# Copyright 2016-2017 DE-CIX Management GmbH
# Author: Tobias Hannaske <tobias.hannaske@de-cix.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import struct

from pbgpp.PCAP.Information import PCAPLayer4Information


class PCAPTCP:
    FIN = 0x01
    SYN = 0x02
    RST = 0x04
    PSH = 0x08
    ACK = 0x10
    URG = 0x20
    ECE = 0x40
    CWR = 0x80

    def __init__(self, payload):
        # Assign variables
        self.payload = payload
        self.ports = None

        self.seq = None
        self.ack = None

        self.flag_fin = False
        self.flag_syn = False
        self.flag_rst = False
        self.flag_psh = False
        self.flag_ack = False
        self.flag_urg = False
        self.flag_ece = False
        self.flag_cwr = False

        self.header_length = None
        self.window_size_value = None
        self.checksum = None
        self.urgent_pointer = None

        # Start parsing
        self.__parse()

    def __parse(self):
        tcp_data = struct.unpack("!HHLL", self.payload[0:12])

        source_port = tcp_data[0]
        destination_port = tcp_data[1]
        self.seq = tcp_data[2]
        self.ack = tcp_data[3]

        self.ports = PCAPLayer4Information(source_port, destination_port)

        length_flags = struct.unpack("!H", self.payload[12:14])[0]
        self.header_length = (length_flags >> 12) * 4

        tcp_flags = struct.unpack("!B", self.payload[13:14])[0]
        if tcp_flags & self.FIN:
            self.flag_fin = True

        if tcp_flags & self.SYN:
            self.flag_syn = True

        if tcp_flags & self.RST:
            self.flag_rst = True

        if tcp_flags & self.PSH:
            self.flag_psh = True

        if tcp_flags & self.ACK:
            self.flag_ack = True

        if tcp_flags & self.URG:
            self.flag_urg = True

        if tcp_flags & self.ECE:
            self.flag_ece = True

        if tcp_flags & self.CWR:
            self.flag_cwr = True

    def get_ports(self):
        return self.ports

    def get_seq(self):
        return self.seq

    def get_ack(self):
        return self.ack

    def get_header_length(self):
        return self.header_length

    def get_window_size_value(self):
        return self.window_size_value

    def get_checksum(self):
        return self.checksum

    def get_urgent_pointer(self):
        return self.urgent_pointer

    def get_tcp_payload(self):
        return self.payload[self.header_length:]



#
# This file is part of PCAP BGP Parser (pbgpp)
#
# Copyright 2016-2017 DE-CIX Management GmbH
# Author: Tobias Hannaske <tobias.hannaske@de-cix.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import struct


class BGPUpdateCommunity:
    def __init__(self, asn, value):
        self.asn = asn
        self.value = value

    def __str__(self):
        if isinstance(self.asn, int) and isinstance(self.value, int):
            return str(self.asn) + ":" + str(self.value)

        elif isinstance(self.asn, bytes) and isinstance(self.value, bytes):
            asn = str(struct.unpack("!H", self.asn)[0])
            value = str(struct.unpack("!H", self.value)[0])

            return asn + ":" + value

        else:
            return None

    def json(self):
        if isinstance(self.asn, int) and isinstance(self.value, int):
            return {
                "asn": self.asn,
                "value": self.value
            }

        elif isinstance(self.asn, bytes) and isinstance(self.value, bytes):
            asn = str(struct.unpack("!H", self.asn)[0])
            value = str(struct.unpack("!H", self.value)[0])

            return {
                "asn": asn,
                "value": value
            }

        else:
            return None



#!/usr/bin/python
#
# This file is part of PCAP BGP Parser (pbgpp)
#
# Copyright 2016-2017 DE-CIX Management GmbH
# Author: Tobias Hannaske <tobias.hannaske@de-cix.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from pbgpp.BGP.Statics import BGPStatics
from pbgpp.BGP.Translation import BGPTranslation
from pbgpp.BGP.Update.Route import BGPRoute
from pbgpp.Output.Exceptions import OutputFormatterError
from pbgpp.Output.Formatter import BGPFormatter


class HumanReadableFormatter(BGPFormatter):
    def __init__(self):
        pass

    def apply(self, message):
        # Example return:
        #
        # [BGPMessage UPDATE] - [123.123.123.123 -> 123.123.123.123]
        # |- IP: 123.123.123.123 -> 123.123.123.123
        # |- MAC: 11:11:11:11:11:11 -> 11:11:11:11:11
        # |- Unix Time: 1412416346.123245123
        # |
        # |- Withdrawn Routes Length: 0
        # |- Total Path Attribute Length: 55
        # |- Path Attributes
        # |--- ORIGIN: IGP
        # |--- AS_PATH: (9498 9430)
        # |--- NEXT_HOP: 80.81.194.250
        # |--- COMMUNITIES: 9498:1 9498:11 9498:91
        # |- NLRI
        # |--- 203.190.42.0/24
        ##

        # Initialize basic return string and PCAP information
        string = "[BGPMessage " + BGPTranslation.message_type(message.type) + "] - " + str(message.length) + " Bytes\n"
        string += self.prefix(0) + "MAC: " + message.pcap_information.get_mac().get_source_string(separated=True) + " -> " + message.pcap_information.get_mac().get_destination_string(separated=True) + "\n"
        string += self.prefix(0) + "IP: " + message.pcap_information.get_ip().get_source_string() + ":" + message.pcap_information.get_ports().get_source_string() + " -> " + message.pcap_information.get_ip().get_destination_string() + ":" + message.pcap_information.get_ports().get_destination_string() + "\n"
        string += self.prefix(0) + "Timestamp: " + message.pcap_information.get_timestmap_utc() + " (" + str(message.pcap_information.get_timestamp()[0]) + "." + str(message.pcap_information.get_timestamp()[1]) + ")\n"

        # Display additional information
        if BGPStatics.MESSAGE_TYPE_KEEPALIVE == message.type:
            pass

        if BGPStatics.MESSAGE_TYPE_OPEN == message.type:
            # --- Divider for PCAP information
            string += self.prefix(-1) + "\n"

            string += self.prefix(0) + "Version: " + str(message.version) + "\n"
            string += self.prefix(0) + "My ASN: " + str(message.asn) + "\n"
            string += self.prefix(0) + "Hold Time: " + str(message.hold_time) + "\n"
            string += self.prefix(0) + "BGP Identifier: " + str(BGPRoute.decimal_ip_to_string(message.identifier)) + "\n"

            # --- Optional Parameters
            string += self.prefix(0) + "Optional Parameters Length: " + str(message.optional_parameter_length) + " Bytes" + "\n"

            # Process optional parameters
            if message.optional_parameter_length > 0:
                string += self.prefix(0) + "Optional Parameters:" + "\n"

                for parameter in message.optional_parameter:
                    if parameter.type == BGPStatics.OPEN_CAPABILITY:
                        string += self.prefix(1) + "Parameter: Capability" + "\n"

                        # Process capabilities
                        for capability in parameter.capability_list:
                            if capability.type is not BGPStatics.CAPABILITY_UNKNOWN:
                                string += self.prefix(2) + BGPTranslation.capability(capability.type) + " (" + str(capability.type) + ")\n"
                            else:
                                string += self.prefix(2) + str(capability) + "\n"

                    elif parameter.type == BGPStatics.OPEN_AUTHENTICATION:
                        string += self.prefix(1) + "Parameter: Authentication" + "\n"
                    elif parameter.type == BGPStatics.OPEN_RESERVED:
                        string += self.prefix(1) + "Parameter: Reserved" + "\n"

        if BGPStatics.MESSAGE_TYPE_UPDATE == message.type:
            # --- Divider for PCAP information
            string += self.prefix(-1) + "\n"

            # --- Update Message Sub-Type
            string += self.prefix(0) + "Update Message Sub-Type: " + BGPTranslation.update_subtype(message.subtype) + "\n"

            # --- Lengths
            string += self.prefix(0) + "Withdrawn Routes Length: " + str(message.withdrawn_routes_length) + " Bytes\n"
            string += self.prefix(0) + "Total Path Attribute Length: " + str(message.path_attributes_length) + " Bytes\n"


            # --- NLRI
            if len(message.nlri) > 0:
                string += self.prefix(0) + "Prefix (NLRI):" + "\n"

                # Process NLRI
                for route in message.nlri:
                    string += self.prefix(1) + str(route) + "\n"

            # --- Path Attributes
            if message.path_attributes_length > 0:

                # Process path attributes
                for attribute in message.path_attributes:
                    string += self.prefix(0) + "Path Attributes:" + "\n"

                    if attribute.type == BGPStatics.UPDATE_ATTRIBUTE_EXTENDED_COMMUNITIES:
                        # Extended Communities must be displayed in another way than other attributes
                        string += self.prefix(1) + BGPTranslation.path_attribute(attribute.type) + ":\n"

                        for community in attribute.extended_communities:
                            string += self.prefix(2) + str(community) + "\n"
                    else:
                        # We got a "normal" path attribute
                        string += self.prefix(1) + BGPTranslation.path_attribute(attribute.type) + ": " + str(attribute) + "\n"

            # --- Withdrawn Routes
            if message.withdrawn_routes_length > 0:
                string += self.prefix(0) + "Withdrawn Routes:" + "\n"

                # Process withdrawn routes
                for route in message.withdrawn_routes:
                    string += self.prefix(1) + str(route) + "\n"

        if BGPStatics.MESSAGE_TYPE_NOTIFICATION == message.type:
            pass

        if BGPStatics.MESSAGE_TYPE_ROUTE_REFRESH == message.type:
            pass

        if BGPStatics.MESSAGE_TYPE_RESERVED == message.type:
            pass

        # Return assembled string plus final line break
        return string + "\n"

    @staticmethod
    def prefix(depth=0):
        if depth < -1:
            raise OutputFormatterError("depth must be bigger than -2.")

        if depth == -1:
            return "|"
        else:
            return "|-" + ("--" * depth) + " "



#
# This file is part of PCAP BGP Parser (pbgpp)
#
# Copyright 2016-2017 DE-CIX Management GmbH
# Author: Tobias Hannaske <tobias.hannaske@de-cix.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import struct

from pbgpp.BGP.Statics import BGPStatics
from pbgpp.BGP.Translation import BGPTranslation
from pbgpp.BGP.Update.PathAttribute import BGPPathAttribute
from pbgpp.BGP.Update.Route import BGPRoute


class PathAttributeNextHop(BGPPathAttribute):
    def __init__(self, payload):
        BGPPathAttribute.__init__(self, payload)
        self.type = BGPStatics.UPDATE_ATTRIBUTE_NEXT_HOP

        # Path attribute specific variable
        self.next_hop = None

        self.__parse()

    def __parse(self):
        try:
            self.parsed = True
            self.error = False

            fields = struct.unpack("!I", self.payload)
            self.next_hop = fields[0]

        except Exception as e:
            self.error = True

    def __str__(self):
        if self.parsed and not self.error:
            return str(BGPRoute.decimal_ip_to_string(self.next_hop))
        else:
            return None

    def json(self):
        return {
            "type": self.type,
            "type_string": BGPTranslation.path_attribute(self.type),
            "error": self.error,
            "next_hop": str(BGPRoute.decimal_ip_to_string(self.next_hop))
        }



#
# This file is part of PCAP BGP Parser (pbgpp)
#
# Copyright 2016-2017 DE-CIX Management GmbH
# Author: Tobias Hannaske <tobias.hannaske@de-cix.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from pbgpp.BGP.Statics import BGPStatics
from pbgpp.BGP.Translation import BGPTranslation


class BGPPathAttribute:
    def __init__(self, payload):
        self.payload = payload

        self.type = None
        self.error = None
        self.parsed = False

    def __str__(self):
        return ""

    def json(self):
        return {
            "type": self.type,
            "type_string": BGPTranslation.path_attribute(self.type),
            "error": self.error,
        }

    @staticmethod
    def factory(attribute_type, payload, attribute_flags):
        # Factory pattern for the path attributes of UPDATE message
        if attribute_type == BGPStatics.UPDATE_ATTRIBUTE_ADVERTISER:
            from pbgpp.BGP.Update.PathAttributes.Advertiser import PathAttributeAdvertiser
            return PathAttributeAdvertiser(payload)

        if attribute_type == BGPStatics.UPDATE_ATTRIBUTE_AGGREGATOR:
            from pbgpp.BGP.Update.PathAttributes.Aggregator import PathAttributeAggregator
            return PathAttributeAggregator(payload)

        if attribute_type == BGPStatics.UPDATE_ATTRIBUTE_AIGP:
            from pbgpp.BGP.Update.PathAttributes.AIGP import PathAttributeAIGP
            return PathAttributeAIGP(payload)

        if attribute_type == BGPStatics.UPDATE_ATTRIBUTE_AS4_AGGREGATOR:
            from pbgpp.BGP.Update.PathAttributes.AS4Aggregator import PathAttributeAS4Aggregator
            return PathAttributeAS4Aggregator(payload)

        if attribute_type == BGPStatics.UPDATE_ATTRIBUTE_AS4_PATH:
            from pbgpp.BGP.Update.PathAttributes.AS4Path import PathAttributeAS4Path
            return PathAttributeAS4Path(payload)

        if attribute_type == BGPStatics.UPDATE_ATTRIBUTE_AS_PATHLIMIT:
            from pbgpp.BGP.Update.PathAttributes.ASPathLimit import PathAttributeASPathLimit
            return PathAttributeASPathLimit(payload)

        if attribute_type == BGPStatics.UPDATE_ATTRIBUTE_AS_PATH:
            from pbgpp.BGP.Update.PathAttributes.ASPath import PathAttributeASPath
            return PathAttributeASPath(payload)

        if attribute_type == BGPStatics.UPDATE_ATTRIBUTE_ATOMIC_AGGREGATE:
            from pbgpp.BGP.Update.PathAttributes.AtomicAggregate import PathAttributeAtomicAggregate
            return PathAttributeAtomicAggregate(payload)

        if attribute_type == BGPStatics.UPDATE_ATTRIBUTE_ATTR_SET:
            from pbgpp.BGP.Update.PathAttributes.AttributeSet import PathAttributeAttributeSet
            return PathAttributeAttributeSet(payload)

        if attribute_type == BGPStatics.UPDATE_ATTRIBUTE_BGP_ENTROPY_LABEL_CAPABILITY:
            from pbgpp.BGP.Update.PathAttributes.BGPEntropyLabelCapability import PathAttributeBGPEntropyLabelCapability
            return PathAttributeBGPEntropyLabelCapability(payload)

        if attribute_type == BGPStatics.UPDATE_ATTRIBUTE_BGP_LS:
            from pbgpp.BGP.Update.PathAttributes.BGPLS import PathAttributeBGPLS
            return PathAttributeBGPLS(payload)

        if attribute_type == BGPStatics.UPDATE_ATTRIBUTE_CLUSTER_LIST:
            from pbgpp.BGP.Update.PathAttributes.ClusterList import PathAttributeClusterList
            return PathAttributeClusterList(payload)

        if attribute_type == BGPStatics.UPDATE_ATTRIBUTE_COMMUNITIES:
            from pbgpp.BGP.Update.PathAttributes.Communities import PathAttributeCommunities
            return PathAttributeCommunities(payload)

        if attribute_type == BGPStatics.UPDATE_ATTRIBUTE_CONNECTOR_ATTRIBUTE:
            from pbgpp.BGP.Update.PathAttributes.ConnectorAttribute import PathAttributeConnectorAttribute
            return PathAttributeConnectorAttribute(payload)

        if attribute_type == BGPStatics.UPDATE_ATTRIBUTE_DPA:
            from pbgpp.BGP.Update.PathAttributes.DPA import PathAttributeDPA
            return PathAttributeDPA(payload)

        if attribute_type == BGPStatics.UPDATE_ATTRIBUTE_EXTENDED_COMMUNITIES:
            from pbgpp.BGP.Update.PathAttributes.ExtendedCommunities import PathAttributeExtendedCommunities
            return PathAttributeExtendedCommunities(payload, attribute_flags)

        if attribute_type == BGPStatics.UPDATE_ATTRIBUTE_IPV6_ADDRESS_SPECIFIC_EXTENDED_COMMUNITY:
            from pbgpp.BGP.Update.PathAttributes.IPv6AddressSpecificExtendedCommunity import PathAttributeIPv6AddressSpecificExtendedCommunitiy
            return PathAttributeIPv6AddressSpecificExtendedCommunitiy(payload)

        if attribute_type == BGPStatics.UPDATE_ATTRIBUTE_LOCAL_PREF:
            from pbgpp.BGP.Update.PathAttributes.LocalPreferences import PathAttributeLocalPreferences
            return PathAttributeLocalPreferences(payload)

        if attribute_type == BGPStatics.UPDATE_ATTRIBUTE_MP_REACH_NLRI:
            from pbgpp.BGP.Update.PathAttributes.MPReachNLRI import PathAttributeMPReachNLRI
            return PathAttributeMPReachNLRI(payload)

        if attribute_type == BGPStatics.UPDATE_ATTRIBUTE_MP_UNREACH_NLRI:
            from pbgpp.BGP.Update.PathAttributes.MPUnReachNLRI import PathAttributeMPUnReachNLRI
            return PathAttributeMPUnReachNLRI(payload)

        if attribute_type == BGPStatics.UPDATE_ATTRIBUTE_MULTI_EXIT_DISC:
            from pbgpp.BGP.Update.PathAttributes.MultipleExitDiscriminator import PathAttributeMultipleExitDiscriminator
            return PathAttributeMultipleExitDiscriminator(payload)

        if attribute_type == BGPStatics.UPDATE_ATTRIBUTE_NEXT_HOP:
            from pbgpp.BGP.Update.PathAttributes.NextHop import PathAttributeNextHop
            return PathAttributeNextHop(payload)

        if attribute_type == BGPStatics.UPDATE_ATTRIBUTE_ORIGIN:
            from pbgpp.BGP.Update.PathAttributes.Origin import PathAttributeOrigin
            return PathAttributeOrigin(payload)

        if attribute_type == BGPStatics.UPDATE_ATTRIBUTE_ORIGINATOR_ID:
            from pbgpp.BGP.Update.PathAttributes.OriginatorID import PathAttributeOriginatorID
            return PathAttributeOriginatorID(payload)

        if attribute_type == BGPStatics.UPDATE_ATTRIBUTE_PE_DISTINGUISHER_LABLES:
            from pbgpp.BGP.Update.PathAttributes.PEDistinguisherLabels import PathAttributePEDistinguisherLabels
            return PathAttributePEDistinguisherLabels(payload)

        if attribute_type == BGPStatics.UPDATE_ATTRIBUTE_PMSI_TUNNEL:
            from pbgpp.BGP.Update.PathAttributes.PMSITunnel import PathAttributePMSITunnel
            return PathAttributePMSITunnel(payload)

        if attribute_type == BGPStatics.UPDATE_ATTRIBUTE_RCID_PATH_CLUSTER_ID:
            from pbgpp.BGP.Update.PathAttributes.PMSITunnel import PathAttributePMSITunnel
            return PathAttributePMSITunnel(payload)

        if attribute_type == BGPStatics.UPDATE_ATTRIBUTE_RESERVED:
            from pbgpp.BGP.Update.PathAttributes.Reserved import PathAttributeReserved
            return PathAttributeReserved(payload)

        if attribute_type == BGPStatics.UPDATE_ATTRIBUTE_RESERVED_DEVELOPMENT:
            from pbgpp.BGP.Update.PathAttributes.ReservedDevelopment import PathAttributeReservedDevelopment
            return PathAttributeReservedDevelopment(payload)

        if attribute_type == BGPStatics.UPDATE_ATTRIBUTE_SAFI_SSA:
            from pbgpp.BGP.Update.PathAttributes.SAFISSA import PathAttributeSAFISSA
            return PathAttributeSAFISSA(payload)

        if attribute_type == BGPStatics.UPDATE_ATTRIBUTE_TRAFFIC_ENGINEERING:
            from pbgpp.BGP.Update.PathAttributes.TrafficEngineering import PathAttributeTrafficEngineering
            return PathAttributeTrafficEngineering(payload)

        if attribute_type == BGPStatics.UPDATE_ATTRIBUTE_TUNNEL_ENCAPSULATION:
            from pbgpp.BGP.Update.PathAttributes.TunnelEncapsulation import PathAttributeTunnelEncapsulation
            return PathAttributeTunnelEncapsulation(payload)

        if attribute_type == BGPStatics.UPDATE_ATTRIBUTE_LARGE_COMMUNITIES:
            from pbgpp.BGP.Update.PathAttributes.LargeCommunities import PathAttributeLargeCommunities
            return PathAttributeLargeCommunities(payload)

        from pbgpp.BGP.Update.PathAttributes.Unknown import PathAttributeUnknown
        return PathAttributeUnknown(payload)



#
# This file is part of PCAP BGP Parser (pbgpp)
#
# Copyright 2016-2017 DE-CIX Management GmbH
# Author: Tobias Hannaske <tobias.hannaske@de-cix.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from pbgpp.BGP.Statics import BGPStatics
from pbgpp.Output.Filter import BGPFilter


class MessageTypeFilter(BGPFilter):
    def __init__(self, values=[]):
        BGPFilter.__init__(self, values)

    def apply(self, message):
        try:
            for value in self.values:
                negated = False

                if value[0:1] == "~":
                    negated = True
                    value = value[1:]

                if value == "RESERVED" or value == "0":
                    check_value = BGPStatics.MESSAGE_TYPE_RESERVED
                elif value == "OPEN" or value == "1":
                    check_value = BGPStatics.MESSAGE_TYPE_OPEN
                elif value == "UPDATE" or value == "2":
                    check_value = BGPStatics.MESSAGE_TYPE_UPDATE
                elif value == "NOTIFICATION" or value == "3":
                    check_value = BGPStatics.MESSAGE_TYPE_NOTIFICATION
                elif value == "KEEPALIVE" or value == "4":
                    check_value = BGPStatics.MESSAGE_TYPE_KEEPALIVE
                elif value == "ROUTE-REFRESH" or value == "ROUTEREFRESH" or value == "5":
                    check_value = BGPStatics.MESSAGE_TYPE_ROUTE_REFRESH
                else:
                    return None

                if not negated and message.type == check_value:
                    # Match on message type
                    return message

                if negated and message.type != check_value:
                    return message

            # Searched value was not found
            return None
        except Exception as e:
            # On error the filtering was not successful (due to wrong fields, etc.)
            return None



#
# This file is part of PCAP BGP Parser (pbgpp)
#
# Copyright 2016-2017 DE-CIX Management GmbH
# Author: Tobias Hannaske <tobias.hannaske@de-cix.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging

from pbgpp.BGP.Statics import BGPStatics


class BGPTranslation:

    @staticmethod
    def boolean(value):
        try:
            if value == True:
                return "True"
            elif value == False:
                return "False"
            else:
                return "Unknown"
        except Exception as e:
            logger = logging.getLogger("pbgpp.BGPTranslation.boolean")
            logger.warning("Was not able to recognize input value for boolean translation.")

            return "Unknown"

    @staticmethod
    def message_type(value):
        try:
            if value == BGPStatics.MESSAGE_TYPE_RESERVED:
                return "RESERVED"
            elif value == BGPStatics.MESSAGE_TYPE_OPEN:
                return "OPEN"
            elif value == BGPStatics.MESSAGE_TYPE_UPDATE:
                return "UPDATE"
            elif value == BGPStatics.MESSAGE_TYPE_NOTIFICATION:
                return "NOTIFICATION"
            elif value == BGPStatics.MESSAGE_TYPE_KEEPALIVE:
                return "KEEPALIVE"
            elif value == BGPStatics.MESSAGE_TYPE_ROUTE_REFRESH:
                return "ROUTE-REFRESH"
            else:
                return "Unknown"
        except Exception as e:
            logger = logging.getLogger("pbgpp.BGPTranslation.message_type")
            logger.warning("Was not able to recognize input value for message type translation.")

            return "Unknown"

    @staticmethod
    def update_subtype(value):
        try:
            if value == BGPStatics.UPDATE_TYPE_NONE:
                return "NONE"
            elif value == BGPStatics.UPDATE_TYPE_WITHDRAWAL:
                return "WITHDRAWAL"
            elif value == BGPStatics.UPDATE_TYPE_ANNOUNCE:
                return "ANNOUNCE"
            elif value == BGPStatics.UPDATE_TYPE_BOTH:
                return "ANNOUNCE/WITHDRAWAL"
            else:
                return "Unknown"
        except Exception as e:
            logger = logging.getLogger("pbgpp.BGPTranslation.update_subtype")
            logger.warning("Was not able to recognize input value for update message subtype translation.")

            return "Unknown"

    @staticmethod
    def open_parameter(value):
        try:
            if value == BGPStatics.OPEN_RESERVED:
                return "RESERVED"
            elif value == BGPStatics.OPEN_AUTHENTICATION:
                return "AUTHENTICATION (Deprecated)"
            elif value == BGPStatics.OPEN_CAPABILITY:
                return "CAPABILITY"
            else:
                return "Unknown"
        except Exception as e:
            logger = logging.getLogger("pbgpp.BGPTranslation.open_parameter")
            logger.warning("Was not able to recognize input value for open message parameter translation.")

            return "Unknown"

    @staticmethod
    def capability(value):
        try:
            if value == BGPStatics.CAPABILITY_RESERVED:
                return "Reserved"
            elif value == BGPStatics.CAPABILITY_MULTIPROTOCOL_EXTENSIONS:
                return "Multiprotocol extensions capability"
            elif value == BGPStatics.CAPABILITY_ROUTE_REFRESH:
                return "Route refresh capability"
            elif value == BGPStatics.CAPABILITY_OUTBOUND_ROUTE_FILTERING:
                return "Outbound route filtering capability"
            elif value == BGPStatics.CAPABILITY_MULTIPLE_ROUTES_TO_DESTINATION:
                return "Multiple routes to a destination capability"
            elif value == BGPStatics.CAPABILITY_EXTENDED_NEXT_HOP_ENCODING:
                return "Extended next hop encoding capability"
            elif value == BGPStatics.CAPABILITY_BGP_EXTENDED:
                return "BGP-Extended capability (Temporary capability - only valid until 2016-09-30)"
            elif value == BGPStatics.CAPABILITY_GRACEFUL_RESTART:
                return "Graceful restart capability"
            elif value == BGPStatics.CAPABILITY_SUPPORT_FOR_FOUR_OCTET_AS:
                return "Support for 4-octet AS number capability"
            elif value == BGPStatics.CAPABILITY_SUPPORT_FOR_DYNAMIC_CAPABILITY:
                return "Support for dynamic capability (capability specific)"
            elif value == BGPStatics.CAPABILITY_MULTISESSION_BGP:
                return "Multisession BGP capability"
            elif value == BGPStatics.CAPABILITY_ADD_PATH:
                return "ADD-PATH capability"
            elif value == BGPStatics.CAPABILITY_ENHANCED_ROUTE_REFRESH:
                return "Enhanced route refresh capability"
            elif value == BGPStatics.CAPABILITY_LONG_LIVED_GRACEFUL_RESTART:
                return "Long-lived graceful restart (LLGR) capability"
            elif value == BGPStatics.CAPABILITY_FQDN:
                return "FQDN capability"
            elif value == BGPStatics.CAPABILITY_ALTERNATIVE_ROUTE_REFRESH:
                return "Route refresh capability (Cisco-legacy)"
            else:
                return "Unknown"
        except Exception as e:
            logger = logging.getLogger("pbgpp.BGPTranslation.capability")
            logger.warning("Was not able to recognize input value for capability translation.")

            return "Unknown"

    @staticmethod
    def origin(value):
        try:
            if value == BGPStatics.ORIGIN_IGP:
                return "IGP"
            elif value == BGPStatics.ORIGIN_EGP:
                return "EGP"
            elif value == BGPStatics.ORIGIN_INCOMPLETE:
                return "INCOMPLETE"
            else:
                return "Unknown"
        except Exception as e:
            logger = logging.getLogger("pbgpp.BGPTranslation.origin")
            logger.warning("Was not able to recognize input value for origin translation.")

            return "Unknown"

    @staticmethod
    def path_attribute(value):
        try:
            if value == BGPStatics.UPDATE_ATTRIBUTE_RESERVED:
                return "Reserved"
            elif value == BGPStatics.UPDATE_ATTRIBUTE_RESERVED_DEVELOPMENT:
                return "Reserved (Development)"
            elif value == BGPStatics.UPDATE_ATTRIBUTE_ORIGIN:
                return "ORIGIN"
            elif value == BGPStatics.UPDATE_ATTRIBUTE_AS_PATH:
                return "AS_PATH"
            elif value == BGPStatics.UPDATE_ATTRIBUTE_NEXT_HOP:
                return "NEXT_HOP"
            elif value == BGPStatics.UPDATE_ATTRIBUTE_MULTI_EXIT_DISC:
                return "MULTI_EXIT_DISC"
            elif value == BGPStatics.UPDATE_ATTRIBUTE_LOCAL_PREF:
                return "LOCAL_PREF"
            elif value == BGPStatics.UPDATE_ATTRIBUTE_ATOMIC_AGGREGATE:
                return "ATOMIC_AGGREGATE"
            elif value == BGPStatics.UPDATE_ATTRIBUTE_AGGREGATOR:
                return "AGGREGATOR"
            elif value == BGPStatics.UPDATE_ATTRIBUTE_COMMUNITIES:
                return "COMMUNITIES"
            elif value == BGPStatics.UPDATE_ATTRIBUTE_ORIGINATOR_ID:
                return "ORIGINATOR_ID"
            elif value == BGPStatics.UPDATE_ATTRIBUTE_CLUSTER_LIST:
                return "CLUSTER_LIST"
            elif value == BGPStatics.UPDATE_ATTRIBUTE_DPA:
                return "DPA"
            elif value == BGPStatics.UPDATE_ATTRIBUTE_ADVERTISER:
                return "ADVERTISER"
            elif value == BGPStatics.UPDATE_ATTRIBUTE_RCID_PATH_CLUSTER_ID:
                return "RCID_PATH / CLUSTER_ID"
            elif value == BGPStatics.UPDATE_ATTRIBUTE_MP_REACH_NLRI:
                return "MP_REACH_NLRI"
            elif value == BGPStatics.UPDATE_ATTRIBUTE_MP_UNREACH_NLRI:
                return "MP_UNREACH_NLRI"
            elif value == BGPStatics.UPDATE_ATTRIBUTE_EXTENDED_COMMUNITIES:
                return "EXTENDED_COMMUNITIES"
            elif value == BGPStatics.UPDATE_ATTRIBUTE_AS4_PATH:
                return "AS4_PATH"
            elif value == BGPStatics.UPDATE_ATTRIBUTE_AS4_AGGREGATOR:
                return "AS4_AGGREGATOR"
            elif value == BGPStatics.UPDATE_ATTRIBUTE_SAFI_SSA:
                return "SAFI Specific Attribute (SSA)"
            elif value == BGPStatics.UPDATE_ATTRIBUTE_CONNECTOR_ATTRIBUTE:
                return "Connector Attribute"
            elif value == BGPStatics.UPDATE_ATTRIBUTE_AS_PATHLIMIT:
                return "AS_PATHLIMIT (deprecated)"
            elif value == BGPStatics.UPDATE_ATTRIBUTE_PMSI_TUNNEL:
                return "PMSI_TUNNEL"
            elif value == BGPStatics.UPDATE_ATTRIBUTE_TUNNEL_ENCAPSULATION:
                return "Tunnel Encapsulation Attribute"
            elif value == BGPStatics.UPDATE_ATTRIBUTE_TRAFFIC_ENGINEERING:
                return "Traffic Engineering"
            elif value == BGPStatics.UPDATE_ATTRIBUTE_IPV6_ADDRESS_SPECIFIC_EXTENDED_COMMUNITY:
                return "IPv6 Address Specific Extended Community"
            elif value == BGPStatics.UPDATE_ATTRIBUTE_AIGP:
                return "AIGP"
            elif value == BGPStatics.UPDATE_ATTRIBUTE_PE_DISTINGUISHER_LABLES:
                return "PE Distinguisher Labels"
            elif value == BGPStatics.UPDATE_ATTRIBUTE_BGP_ENTROPY_LABEL_CAPABILITY:
                return "BGP Entropy Label Capability Attribute"
            elif value == BGPStatics.UPDATE_ATTRIBUTE_BGP_LS:
                return "BGP-LS Attribute"
            elif value == BGPStatics.UPDATE_ATTRIBUTE_ATTR_SET:
                return "ATTR_SET"
            elif value == BGPStatics.UPDATE_ATTRIBUTE_LARGE_COMMUNITIES:
                return "LARGE_COMMUNITIES"
            else:
                return "Unknown"
        except Exception as e:
            logger = logging.getLogger("pbgpp.BGPTranslation.path_attribute")
            logger.warning("Was not able to recognize input value for path attribute translation.")

            return "Unknown"

    @staticmethod
    def extended_community(type, subtype):
        try:
            type_string = BGPTranslation.extended_community_type(type)

            if type == BGPStatics.EXT_COMMUNITY_T_TWO_OCTET_AS_SPECIFIC:
                sub_type_string = BGPTranslation.extended_community_t_two_octet_as_subtype(subtype)
            elif type == BGPStatics.EXT_COMMUNITY_T_IPV4_ADDRESS_SPECIFIC:
                sub_type_string = BGPTranslation.extended_community_t_ipv4_address_subtype(subtype)
            elif type == BGPStatics.EXT_COMMUNITY_T_FOUR_OCTET_AS_SPECIFIC:
                sub_type_string = BGPTranslation.extended_community_t_four_octet_as_subtype(subtype)
            elif type == BGPStatics.EXT_COMMUNITY_T_OPAQUE:
                sub_type_string = BGPTranslation.extended_community_t_opaque(subtype)
            elif type == BGPStatics.EXT_COMMUNITY_T_QOS_MARKING:
                sub_type_string = ""
            elif type == BGPStatics.EXT_COMMUNITY_T_COS_CAPABILITY:
                sub_type_string = ""
            elif type == BGPStatics.EXT_COMMUNITY_T_EVPN:
                sub_type_string = BGPTranslation.extended_community_evpn_subtype(subtype)
            elif type == BGPStatics.EXT_COMMUNITY_T_FLOW_SPEC_REDIRECT_MIRROR_TO_IP_NEXT_HOP:
                sub_type_string = ""
            elif type == BGPStatics.EXT_COMMUNITY_T_GENERIC_EXPERIMENTAL_USE:
                sub_type_string = BGPTranslation.extended_community_t_generic_experimental(subtype)
            elif type == BGPStatics.EXT_COMMUNITY_T_GENERIC_EXPERIMENTAL_USE_PART2:
                sub_type_string = BGPTranslation.extended_community_t_generic_experimental_part2(subtype)
            elif type == BGPStatics.EXT_COMMUNITY_T_GENERIC_EXPERIMENTAL_USE_PART3:
                sub_type_string = BGPTranslation.extended_community_t_generic_experimental_part3(subtype)
            elif type == BGPStatics.EXT_COMMUNITY_NT_TWO_OCTET_AS_SPECIFIC:
                sub_type_string = BGPTranslation.extended_community_nt_two_octet_as_subtype(subtype)
            elif type == BGPStatics.EXT_COMMUNITY_NT_IPV4_ADDRESS_SPECIFIC:
                sub_type_string = BGPTranslation.extended_community_nt_ipv4_address_subtype(subtype)
            elif type == BGPStatics.EXT_COMMUNITY_NT_FOUR_OCTET_AS_SPECIFIC:
                sub_type_string = BGPTranslation.extended_community_nt_four_octet_as_subtype(subtype)
            elif type == BGPStatics.EXT_COMMUNITY_NT_OPAQUE:
                sub_type_string = BGPTranslation.extended_community_nt_opaque(subtype)
            elif type == BGPStatics.EXT_COMMUNITY_NT_QOS_MARKING:
                sub_type_string = ""
            else:
                sub_type_string = "Unknown Sub-Type"

            return type_string + " " + sub_type_string
        except Exception as e:
            logger = logging.getLogger("pbgpp.BGPTranslation.extended_community")
            logger.warning("Was not able to recognize input value for extended community translation.")

            return "Unknown Sub-Type"

    @staticmethod
    def extended_community_type(value):
        try:
            if value == BGPStatics.EXT_COMMUNITY_NT_TWO_OCTET_AS_SPECIFIC:
                return "Non-Transitive Two-Octet AS"
            elif value == BGPStatics.EXT_COMMUNITY_NT_IPV4_ADDRESS_SPECIFIC:
                return "Non-Transitive IPv4"
            elif value == BGPStatics.EXT_COMMUNITY_NT_FOUR_OCTET_AS_SPECIFIC:
                return "Non-Transitive Four-Octet AS"
            elif value == BGPStatics.EXT_COMMUNITY_NT_OPAQUE:
                return "Non-Transitive Opaque"
            elif value == BGPStatics.EXT_COMMUNITY_NT_QOS_MARKING:
                return "QoS Marking"
            elif value == BGPStatics.EXT_COMMUNITY_T_TWO_OCTET_AS_SPECIFIC:
                return "Transitive Two-Octet AS"
            elif value == BGPStatics.EXT_COMMUNITY_T_IPV4_ADDRESS_SPECIFIC:
                return "Transitive IPv4"
            elif value == BGPStatics.EXT_COMMUNITY_T_FOUR_OCTET_AS_SPECIFIC:
                return "Transitive Four-Octet AS"
            elif value == BGPStatics.EXT_COMMUNITY_T_OPAQUE:
                return "Transitive Opaque"
            elif value == BGPStatics.EXT_COMMUNITY_T_QOS_MARKING:
                return "QoS Marking"
            elif value == BGPStatics.EXT_COMMUNITY_T_COS_CAPABILITY:
                return "CoS Capability"
            elif value == BGPStatics.EXT_COMMUNITY_T_EVPN:
                return "EVPN"
            elif value == BGPStatics.EXT_COMMUNITY_T_FLOW_SPEC_REDIRECT_MIRROR_TO_IP_NEXT_HOP:
                return "Flow spec redirect/mirror to IP next-hop"
            elif value == BGPStatics.EXT_COMMUNITY_T_GENERIC_EXPERIMENTAL_USE:
                return "Generic Transitive Experimental Use"
            elif value == BGPStatics.EXT_COMMUNITY_T_GENERIC_EXPERIMENTAL_USE_PART2:
                return "Generic Transitive Experimental Use (Part 2)"
            elif value == BGPStatics.EXT_COMMUNITY_T_GENERIC_EXPERIMENTAL_USE_PART3:
                return "Generic Transitive Experimental Use (Part 3)"
            else:
                return "Unknown"
        except Exception as e:
            logger = logging.getLogger("pbgpp.BGPTranslation.extended_community_type")
            logger.warning("Was not able to recognize input value for extended community type translation.")

            return "Unknown"

    @staticmethod
    def extended_community_subtype(type, value):
        try:
            if type == BGPStatics.EXT_COMMUNITY_T_TWO_OCTET_AS_SPECIFIC:
                sub_type_string = BGPTranslation.extended_community_t_two_octet_as_subtype(value)
            elif type == BGPStatics.EXT_COMMUNITY_T_IPV4_ADDRESS_SPECIFIC:
                sub_type_string = BGPTranslation.extended_community_t_ipv4_address_subtype(value)
            elif type == BGPStatics.EXT_COMMUNITY_T_FOUR_OCTET_AS_SPECIFIC:
                sub_type_string = BGPTranslation.extended_community_t_four_octet_as_subtype(value)
            elif type == BGPStatics.EXT_COMMUNITY_T_OPAQUE:
                sub_type_string = BGPTranslation.extended_community_t_opaque(value)
            elif type == BGPStatics.EXT_COMMUNITY_T_QOS_MARKING:
                sub_type_string = ""
            elif type == BGPStatics.EXT_COMMUNITY_T_COS_CAPABILITY:
                sub_type_string = ""
            elif type == BGPStatics.EXT_COMMUNITY_T_EVPN:
                sub_type_string = BGPTranslation.extended_community_evpn_subtype(value)
            elif type == BGPStatics.EXT_COMMUNITY_T_FLOW_SPEC_REDIRECT_MIRROR_TO_IP_NEXT_HOP:
                sub_type_string = ""
            elif type == BGPStatics.EXT_COMMUNITY_T_GENERIC_EXPERIMENTAL_USE:
                sub_type_string = BGPTranslation.extended_community_t_generic_experimental(value)
            elif type == BGPStatics.EXT_COMMUNITY_T_GENERIC_EXPERIMENTAL_USE_PART2:
                sub_type_string = BGPTranslation.extended_community_t_generic_experimental_part2(value)
            elif type == BGPStatics.EXT_COMMUNITY_T_GENERIC_EXPERIMENTAL_USE_PART3:
                sub_type_string = BGPTranslation.extended_community_t_generic_experimental_part3(value)
            elif type == BGPStatics.EXT_COMMUNITY_NT_TWO_OCTET_AS_SPECIFIC:
                sub_type_string = BGPTranslation.extended_community_nt_two_octet_as_subtype(value)
            elif type == BGPStatics.EXT_COMMUNITY_NT_IPV4_ADDRESS_SPECIFIC:
                sub_type_string = BGPTranslation.extended_community_nt_ipv4_address_subtype(value)
            elif type == BGPStatics.EXT_COMMUNITY_NT_FOUR_OCTET_AS_SPECIFIC:
                sub_type_string = BGPTranslation.extended_community_nt_four_octet_as_subtype(value)
            elif type == BGPStatics.EXT_COMMUNITY_NT_OPAQUE:
                sub_type_string = BGPTranslation.extended_community_nt_opaque(value)
            elif type == BGPStatics.EXT_COMMUNITY_NT_QOS_MARKING:
                sub_type_string = ""
            else:
                sub_type_string = "Unknown Sub-Type"

            return sub_type_string
        except Exception as e:
            logger = logging.getLogger("pbgpp.BGPTranslation.extended_community_subtype")
            logger.warning("Was not able to recognize input value for extended community subtype translation.")

            return "Unknown Sub-Type"

    @staticmethod
    def extended_community_evpn_subtype(value):
        try:
            if value == BGPStatics.EXT_COMMUNITY_EVPN_MAC_MOBILITY:
                return "MAC Mobility"
            elif value == BGPStatics.EXT_COMMUNITY_EVPN_ESI_LABEL:
                return "ESI Label"
            elif value == BGPStatics.EXT_COMMUNITY_EVPN_ES_IMPORT_ROUTE_TARGET:
                return "ES-Import Route Target"
            elif value == BGPStatics.EXT_COMMUNITY_EVPN_ROUTERS_MAC:
                return "EVPN Router's MAC Extended Community"
            elif value == BGPStatics.EXT_COMMUNITY_EVPN_LAYER2:
                return "Layer 2 Extended Community"
            elif value == BGPStatics.EXT_COMMUNITY_EVPN_ETREE:
                return "E-TREE Extended Community"
            elif value == BGPStatics.EXT_COMMUNITY_EVPN_DF_ELECTION:
                return "DF Election Extended Community"
            elif value == BGPStatics.EXT_COMMUNITY_EVPN_ISID:
                return "I-SID Extended Community"
            else:
                return "Unknown"
        except Exception as e:
            logger = logging.getLogger("pbgpp.BGPTranslation.extended_community_evpn_subtype")
            logger.warning("Was not able to recognize input value for extended community EVPN subtype translation.")

            return "Unknown"

    @staticmethod
    def extended_community_t_two_octet_as_subtype(value):
        try:
            if value == BGPStatics.EXT_COMMUNITY_T_TWO_OCTET_AS_ROUTE_TARGET:
                return "Route Target"
            elif value == BGPStatics.EXT_COMMUNITY_T_TWO_OCTET_AS_ROUTE_ORIGIN:
                return "Route Origin"
            elif value == BGPStatics.EXT_COMMUNITY_T_TWO_OCTET_AS_OSPF_DOMAIN_IDENTIFIER:
                return "OSPF Domain Identifier"
            elif value == BGPStatics.EXT_COMMUNITY_T_TWO_OCTET_AS_BGP_DATA_COLLECTION:
                return "BGP Data Collection"
            elif value == BGPStatics.EXT_COMMUNITY_T_TWO_OCTET_AS_SOURCE_AS:
                return "Source AS"
            elif value == BGPStatics.EXT_COMMUNITY_T_TWO_OCTET_AS_L2VPN_IDENTIFIER:
                return "L2VPN Identifier"
            elif value == BGPStatics.EXT_COMMUNITY_T_TWO_OCTET_AS_CISCO_VPN_DISTINGUISHER:
                return "Cisco VPN-Distinguisher"
            elif value == BGPStatics.EXT_COMMUNITY_T_TWO_OCTET_AS_ROUTE_TARGET_RECORD:
                return "Route-Target Record"
            else:
                return "Unknown"
        except Exception as e:
            logger = logging.getLogger("pbgpp.BGPTranslation.extended_community_t_two_octet_as_subtype")
            logger.warning("Was not able to recognize input value for type translation.")

            return "Unknown"

    @staticmethod
    def extended_community_nt_two_octet_as_subtype(value):
        try:
            if value == BGPStatics.EXT_COMMUNITY_NT_TWO_OCTET_AS_LINK_BANDWIDTH:
                return "Link Bandwidth Extended Community"
            elif value == BGPStatics.EXT_COMMUNITY_NT_TWO_OCTET_AS_VIRTUAL_NETWORK_IDENTIFIER:
                return "Virtual-Network Identifier Extended Community"
            else:
                return "Unknown"
        except Exception as e:
            logger = logging.getLogger("pbgpp.BGPTranslation.extended_community_nt_two_octet_as_subtype")
            logger.warning("Was not able to recognize input value for type translation.")

            return "Unknown"

    @staticmethod
    def extended_community_t_four_octet_as_subtype(value):
        try:
            if value == BGPStatics.EXT_COMMUNITY_T_FOUR_OCTET_AS_ROUTE_TARGET:
                return "Route Target"
            elif value == BGPStatics.EXT_COMMUNITY_T_FOUR_OCTET_AS_ROUTE_ORIGIN:
                return "Route Origin"
            elif value == BGPStatics.EXT_COMMUNITY_T_FOUR_OCTET_AS_GENERIC:
                return "Generic"
            elif value == BGPStatics.EXT_COMMUNITY_T_FOUR_OCTET_AS_OSPF_DOMAIN_IDENTIFIER:
                return "OSPF Domain Identifier"
            elif value == BGPStatics.EXT_COMMUNITY_T_FOUR_OCTET_AS_BGP_DATA_COLLECTION:
                return "BGP Data Collection"
            elif value == BGPStatics.EXT_COMMUNITY_T_FOUR_OCTET_AS_SOURCE_AS:
                return "Source AS"
            elif value == BGPStatics.EXT_COMMUNITY_T_FOUR_OCTET_AS_CISCO_VPN_IDENTIFIER:
                return "Cisco VPN Identifier"
            elif value == BGPStatics.EXT_COMMUNITY_T_FOUR_OCTET_AS_ROUTE_TARGET_RECORD:
                return "Route-Target Record"
            else:
                return "Unknown"
        except Exception as e:
            logger = logging.getLogger("pbgpp.BGPTranslation.extended_community_t_four_octet_as_subtype")
            logger.warning("Was not able to recognize input value for type translation.")

            return "Unknown"

    @staticmethod
    def extended_community_nt_four_octet_as_subtype(value):
        try:
            if value == BGPStatics.EXT_COMMUNITY_NT_FOUR_OCTET_AS_GENERIC:
                return "Generic"
            else:
                return "Unknown"
        except Exception as e:
            logger = logging.getLogger("pbgpp.BGPTranslation.extended_community_nt_four_octet_as_subtype")
            logger.warning("Was not able to recognize input value for type translation.")

            return "Unknown"

    @staticmethod
    def extended_community_t_ipv4_address_subtype(value):
        try:
            if value == BGPStatics.EXT_COMMUNITY_T_IPV4_ADDRESS_ROUTE_TARGET:
                return "Route Target"
            elif value == BGPStatics.EXT_COMMUNITY_T_IPV4_ADDRESS_ROUTE_ORIGIN:
                return "Route Origin"
            elif value == BGPStatics.EXT_COMMUNITY_T_IPV4_ADDRESS_OSPF_DOMAIN_IDENTIFIER:
                return "OSPF Domain Identifier"
            elif value == BGPStatics.EXT_COMMUNITY_T_IPV4_ADDRESS_OSPF_ROUTE_ID:
                return "OSPF Route ID"
            elif value == BGPStatics.EXT_COMMUNITY_T_IPV4_ADDRESS_L2VPN_IDENTIFIER:
                return "L2VPN Identifier"
            elif value == BGPStatics.EXT_COMMUNITY_T_IPV4_ADDRESS_VRF_ROUTE_IMPORT:
                return "VRF Route Import"
            elif value == BGPStatics.EXT_COMMUNITY_T_IPV4_ADDRESS_FLOW_SPEC_REDIRECT_TO_IPV4:
                return "Flow-spec Redirect to IPv4"
            elif value == BGPStatics.EXT_COMMUNITY_T_IPV4_ADDRESS_CISCO_VPN_DISTINGUISHER:
                return "Cisco VPN-Distinguisher"
            elif value == BGPStatics.EXT_COMMUNITY_T_IPV4_ADDRESS_INTER_AREA_P2MP_SEGMENTED_NEXT_HOP:
                return "Inter-Area P2MP Segmented Next-Hop"
            elif value == BGPStatics.EXT_COMMUNITY_T_IPV4_ADDRESS_ROUTE_TARGET_RECORD:
                return "Route-Target-Record"
            else:
                return "Unknown"
        except Exception as e:
            logger = logging.getLogger("pbgpp.BGPTranslation.extended_community_t_ipv4_address_subtype")
            logger.warning("Was not able to recognize input value for type translation.")

            return "Unknown"

    @staticmethod
    def extended_community_nt_ipv4_address_subtype(value):
        try:
            return "Unknown"
        except Exception as e:
            logger = logging.getLogger("pbgpp.BGPTranslation.extended_community_nt_ipv4_address_subtype")
            logger.warning("Was not able to recognize input value for type translation.")

            return "Unknown"

    @staticmethod
    def extended_community_t_opaque(value):
        try:
            if value == BGPStatics.EXT_COMMUNITY_T_OPAQUE_COST_COMMUNITY:
                return "Cost Community"
            elif value == BGPStatics.EXT_COMMUNITY_T_OPAQUE_CP_ORF:
                return "CP-ORF"
            elif value == BGPStatics.EXT_COMMUNITY_T_OPAQUE_EXTRANET_SOURCE:
                return "Extranet Source Extended Community"
            elif value == BGPStatics.EXT_COMMUNITY_T_OPAQUE_EXTRANET_SEPARATION:
                return "Extranet Separation Extended Community"
            elif value == BGPStatics.EXT_COMMUNITY_T_OPAQUE_OSPF_ROUTE_TYPE:
                return "OSPF Route Type"
            elif value == BGPStatics.EXT_COMMUNITY_T_OPAQUE_ADDITIONAL_PMSI_TUNNEL_ATTRIBUTE_FLAGS:
                return "Additional PMSI Tunnel Attribute Flags"
            elif value == BGPStatics.EXT_COMMUNITY_T_OPAQUE_COLOR_EXTENDED:
                return "Color Extended Community"
            elif value == BGPStatics.EXT_COMMUNITY_T_OPAQUE_ENCAPSULATION:
                return "Encapsulation Extended Community"
            elif value == BGPStatics.EXT_COMMUNITY_T_OPAQUE_DEFAULT_GATEWAY:
                return "Default Gateway"
            elif value == BGPStatics.EXT_COMMUNITY_T_OPAQUE_PPMP_LABEL:
                return "Point-to-Point-to-Multipoint (PPMP) Label"
            elif value == BGPStatics.EXT_COMMUNITY_T_OPAQUE_CONSISTENT_HASH_SORT_ORDER:
                return "Consistent Hash Sort Order"
            else:
                return "Unknown"
        except Exception as e:
            logger = logging.getLogger("pbgpp.BGPTranslation.extended_community_t_opaque")
            logger.warning("Was not able to recognize input value for type translation.")

            return "Unknown"

    @staticmethod
    def extended_community_nt_opaque(value):
        try:
            if value == BGPStatics.EXT_COMMUNITY_NT_OPAQUE_COST_COMMUNITY:
                return "Cost Community"
            elif value == BGPStatics.EXT_COMMUNITY_NT_OPAQUE_BGP_ORIGIN_VALIDATION_STATE:
                return "BGP Origin Validation State"
            else:
                return "Unknown"
        except Exception as e:
            logger = logging.getLogger("pbgpp.BGPTranslation.extended_community_nt_opaque")
            logger.warning("Was not able to recognize input value for type translation.")

            return "Unknown"

    @staticmethod
    def extended_community_t_generic_experimental(value):
        try:
            if value == BGPStatics.EXT_COMMUNITY_T_GENERIC_EXPERIMENTAL_OSPF_ROUTE_TYPE:
                return "OSPF Route Type"
            elif value == BGPStatics.EXT_COMMUNITY_T_GENERIC_EXPERIMENTAL_OSPF_ROUTER_ID:
                return "OSPF Router ID"
            elif value == BGPStatics.EXT_COMMUNITY_T_GENERIC_EXPERIMENTAL_OSPF_DOMAIN_IDENTIFIER:
                return "OSPF Domain Identifier"
            elif value == BGPStatics.EXT_COMMUNITY_T_GENERIC_EXPERIMENTAL_FLOW_SPEC_TRAFFIC_RATE:
                return "Flow spec traffic-rate"
            elif value == BGPStatics.EXT_COMMUNITY_T_GENERIC_EXPERIMENTAL_FLOW_SPEC_TRAFFIC_ACTION:
                return "Flow spec traffic-action"
            elif value == BGPStatics.EXT_COMMUNITY_T_GENERIC_EXPERIMENTAL_FLOW_SEPC_REDIRECT_AS_2BYTE_FORMAT:
                return "Flow spec redirect AS-2byte format"
            elif value == BGPStatics.EXT_COMMUNITY_T_GENERIC_EXPERIMENTAL_FLOW_SPEC_TRAFFIC_REMARKING:
                return "Flow spec traffic-remarking"
            elif value == BGPStatics.EXT_COMMUNITY_T_GENERIC_EXPERIMENTAL_LAYER2_INFO:
                return "Layer2 Info Extended Community"
            elif value == BGPStatics.EXT_COMMUNITY_T_GENERIC_EXPERIMENTAL_E_TREE_INFO:
                return "E-Tree Info"
            else:
                return "Unknown"
        except Exception as e:
            logger = logging.getLogger("pbgpp.BGPTranslation.extended_community_t_generic_experimental")
            logger.warning("Was not able to recognize input value for type translation.")

            return "Unknown"

    @staticmethod
    def extended_community_t_generic_experimental_part2(value):
        try:
            if value == BGPStatics.EXT_COMMUNITY_T_GENERIC_EXPERIMENTAL_PART2_FLOW_SPEC_REDIRECT_IPV4_FORMAT:
                return "Flow spec redirect IPv4 format"
            else:
                return "Unknown"
        except Exception as e:
            logger = logging.getLogger("pbgpp.BGPTranslation.extended_community_t_generic_experimental_part2")
            logger.warning("Was not able to recognize input value for type translation.")

            return "Unknown"

    @staticmethod
    def extended_community_t_generic_experimental_part3(value):
        try:
            if value == BGPStatics.EXT_COMMUNITY_T_GENERIC_EXPERIMENTAL_PART3_FLOW_SPEC_AS_4BYTE_FORMAT:
                return "Flow spec redirect AS-4byte format"
            else:
                return "Unknown"
        except Exception as e:
            logger = logging.getLogger("pbgpp.BGPTranslation.extended_community_t_generic_experimental_part3")
            logger.warning("Was not able to recognize input value for type translation.")

            return "Unknown"

    @staticmethod
    def path_segment_type(value):
        try:
            if value == BGPStatics.AS_PATH_SEGMENT_SEQUENCE:
                return "Sequence"
            elif value == BGPStatics.AS_PATH_SEGMENT_SET:
                return "Set"
            else:
                return "Unknown"
        except Exception as e:
            logger = logging.getLogger("pbgpp.BGPTranslation.path_segment_type")
            logger.warning("Was not able to recognize input value for path segment type translation.")

            return "Unknown"



#
# This file is part of PCAP BGP Parser (pbgpp)
#
# Copyright 2016-2017 DE-CIX Management GmbH
# Author: Tobias Hannaske <tobias.hannaske@de-cix.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging
import re

from pbgpp.BGP.Exceptions import BGPPacketError, BGPPacketHasNoMessagesError, BGPMessageFactoryError, BGPError
from pbgpp.BGP.Message import BGPMessage
from pbgpp.PCAP.Information import PCAPInformation


class BGPPacket:
    def __init__(self, payload, pcap_information):
        # Assign payload and pcap information
        self.payload = payload
        self.pcap_information = pcap_information

        self.message_list = []
        self.__parsed = False
        self.__iteration_position = 0

        # Typecheck pcap information
        if not isinstance(self.pcap_information, PCAPInformation):
            raise BGPPacketError("pcap_information needs to be an object of PCAPInformation class")

        # Parse packet
        self.__parse()

    def __eq__(self, other):
        if isinstance(other, BGPPacket):
            # Compare raw payload to equate both objects
            return self.payload == other.payload
        else:
            return NotImplemented

    def __len__(self):
        # Return the length of raw payload (== the packet length in Bytes)
        return len(self.payload)

    def __iter__(self):
        return self

    def __next__(self):
        try:
            # Return next message
            result = self.message_list[self.__iteration_position]
        except IndexError:
            # No more messages: Raise StopIteration
            raise StopIteration

        # Current iteration position += 1
        self.__iteration_position += 1
        return result

    def __str__(self):
        if self.__parsed:
            # Display amount of BGP messages that are contained in this single TCP packet and their whole length
            return "<BGPPacket messages={0} length={1}>".format(len(self.message_list), len(self.payload))
        else:
            return "<BGPPacket messages=UNKNOWN length=UNKNOWN>"

    def __parse(self):
        logger = logging.getLogger("pbgpp.BGPPacket.__parse")

        # Split the byte string by BGP's magic marker and filter the empty matches from result list
        messages = re.split(b'(?:\xff){16}', self.payload)
        messages = list(filter(None, messages))  # Fastest solution for filtering

        # Check if the payload starts with the magic marker. If not we have to trash that bytes as it's
        # a result of a missing TCP stream reassembly
        if not self.payload[0:16] == b'\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff':
            messages.pop(0)

        # Check for empty list (in this case we have a malformed/non-BGP packet)
        if len(messages) == 0:
            raise BGPPacketHasNoMessagesError("parsed packet didn't contain any BGP messages")

        # Now iterate through the found messages ...
        for m in messages:
            try:
                # ... and add them to the message list of packet object using a message factory pattern
                self.add_message(BGPMessage.factory(m, self.pcap_information))
            except BGPMessageFactoryError as f:
                # This exception can be raised when no valid message type could be found
                # It's a common exception when there is a malformed packet - therefore: log it as INFO
                logger.info("BGPMessageFactoryError was raised due to unknown message type during initial packet parsing.")
            except BGPPacketError as p:
                # This exception can be raised when system tries to add non-BGPMessage object to message list
                # Uncommon to happen - Better raise an error
                logger.error("Tried to add a non-BGPMessage object to message list.")
            except BGPError as e:
                # A lot of other things could go wrong - fall back to a warning message
                logger.warning("Unspecified BGPError raised during initial packet parsing.")

        self.__parsed = True

    def add_message(self, message):
        try:
            # When trying to add non-BGPMessage object raise new BGPPacketException
            if not isinstance(message, BGPMessage):
                raise BGPPacketError("tried to add non-BGPMessage object to message list")
            else:
                self.message_list.append(message)
                return True
        except Exception as e:
            # There should be no case where this Exception should be raised
            return False

    def remove_message(self, message):
        try:
            # Remove message oject from message list
            if message in self.message_list:
                self.message_list.remove(message)
                return True
            else:
                return False
        except Exception as e:
            # There should be no case where this Exception should be raised
            return False

    def get_pcap_information(self):
        try:
            # Return the stored PCAP information if they have been set, yet
            if self.pcap_information is not None:
                return self.pcap_information
            else:
                return False
        except Exception as e:
            return False

    def get_message_list(self):
        try:
            return self.message_list
        except Exception as e:
            # There should be no case where this Exception should be raised
            return False



#
# This file is part of PCAP BGP Parser (pbgpp)
#
# Copyright 2016-2017 DE-CIX Management GmbH
# Author: Tobias Hannaske <tobias.hannaske@de-cix.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


class BGPStatics:
    # ** BGPMessage **
    # Message types according to RFC 4271
    ##
    MESSAGE_TYPE_RESERVED = 0
    MESSAGE_TYPE_OPEN = 1
    MESSAGE_TYPE_UPDATE = 2
    MESSAGE_TYPE_NOTIFICATION = 3
    MESSAGE_TYPE_KEEPALIVE = 4
    MESSAGE_TYPE_ROUTE_REFRESH = 5

    # ** BGPUpdate sub-types **
    # Internal Usage Only
    ##
    UPDATE_TYPE_NONE = 0x0
    UPDATE_TYPE_ANNOUNCE = 0x1
    UPDATE_TYPE_WITHDRAWAL = 0x2
    UPDATE_TYPE_BOTH = 0x3

    # ** BGPOpen **
    # Optional parameters according to RFC 5492 and RFC 4271
    # Unassigned: 3-255
    ##
    OPEN_RESERVED = 0
    OPEN_AUTHENTICATION = 1 # (Deprecated)
    OPEN_CAPABILITY = 2

    # ** BGPOpenOptionalParameterCapability sub-codes **
    ##
    CAPABILITY_UNKNOWN = -1
    CAPABILITY_RESERVED = 0
    CAPABILITY_MULTIPROTOCOL_EXTENSIONS = 1
    CAPABILITY_ROUTE_REFRESH = 2
    CAPABILITY_OUTBOUND_ROUTE_FILTERING = 3
    CAPABILITY_MULTIPLE_ROUTES_TO_DESTINATION = 4
    CAPABILITY_EXTENDED_NEXT_HOP_ENCODING = 5
    CAPABILITY_BGP_EXTENDED = 6
    CAPABILITY_GRACEFUL_RESTART = 64
    CAPABILITY_SUPPORT_FOR_FOUR_OCTET_AS = 65
    CAPABILITY_SUPPORT_FOR_DYNAMIC_CAPABILITY = 67
    CAPABILITY_MULTISESSION_BGP = 68
    CAPABILITY_ADD_PATH = 69
    CAPABILITY_ENHANCED_ROUTE_REFRESH = 70
    CAPABILITY_LONG_LIVED_GRACEFUL_RESTART = 71
    CAPABILITY_FQDN = 73
    CAPABILITY_ALTERNATIVE_ROUTE_REFRESH = 128

    # ** BGPUpdate **
    # Path attributes according to RFC 4271
    # Unassigned: 30-39 && 41-127 && 129-254
    # Temporary: 40 (Until 2016-09-30)
    ##
    UPDATE_ATTRIBUTE_RESERVED = 0
    UPDATE_ATTRIBUTE_ORIGIN = 1
    UPDATE_ATTRIBUTE_AS_PATH = 2
    UPDATE_ATTRIBUTE_NEXT_HOP = 3
    UPDATE_ATTRIBUTE_MULTI_EXIT_DISC = 4
    UPDATE_ATTRIBUTE_LOCAL_PREF = 5
    UPDATE_ATTRIBUTE_ATOMIC_AGGREGATE = 6
    UPDATE_ATTRIBUTE_AGGREGATOR = 7
    UPDATE_ATTRIBUTE_COMMUNITIES = 8
    UPDATE_ATTRIBUTE_ORIGINATOR_ID = 9
    UPDATE_ATTRIBUTE_CLUSTER_LIST = 10
    UPDATE_ATTRIBUTE_DPA = 11 # (Deprecated)
    UPDATE_ATTRIBUTE_ADVERTISER = 12 # (Historic) (Depcrecated)
    UPDATE_ATTRIBUTE_RCID_PATH_CLUSTER_ID = 13 # (Historic) (Deprecated)
    UPDATE_ATTRIBUTE_MP_REACH_NLRI = 14
    UPDATE_ATTRIBUTE_MP_UNREACH_NLRI = 15
    UPDATE_ATTRIBUTE_EXTENDED_COMMUNITIES = 16
    UPDATE_ATTRIBUTE_AS4_PATH = 17
    UPDATE_ATTRIBUTE_AS4_AGGREGATOR = 18
    UPDATE_ATTRIBUTE_SAFI_SSA = 19 # (Deprecated)
    UPDATE_ATTRIBUTE_CONNECTOR_ATTRIBUTE = 20 # (Deprecated)
    UPDATE_ATTRIBUTE_AS_PATHLIMIT = 21 # (Deprecated)
    UPDATE_ATTRIBUTE_PMSI_TUNNEL = 22
    UPDATE_ATTRIBUTE_TUNNEL_ENCAPSULATION = 23
    UPDATE_ATTRIBUTE_TRAFFIC_ENGINEERING = 24
    UPDATE_ATTRIBUTE_IPV6_ADDRESS_SPECIFIC_EXTENDED_COMMUNITY = 25
    UPDATE_ATTRIBUTE_AIGP = 26
    UPDATE_ATTRIBUTE_PE_DISTINGUISHER_LABLES = 27
    UPDATE_ATTRIBUTE_BGP_ENTROPY_LABEL_CAPABILITY = 28 # (Deprecated)
    UPDATE_ATTRIBUTE_BGP_LS = 29
    UPDATE_ATTRIBUTE_LARGE_COMMUNITIES = 32
    UPDATE_ATTRIBUTE_ATTR_SET = 128
    UPDATE_ATTRIBUTE_RESERVED_DEVELOPMENT = 255

    # ** BGPNotification **
    # BGP Error (Notification) Codes according to RFC 4271 and RFC 7313
    # Unassigned: 8-255
    ##
    NOTIFICATION_ERROR_RESERVED = 0
    NOTIFICATION_ERROR_MESSAGE_HEADER = 1
    NOTIFICATION_ERROR_OPEN_MESSAGE = 2
    NOTIFICATION_ERROR_UPDATE_MESSAGE = 3
    NOTIFICATION_ERROR_HOLD_TIMER_EXPIRED = 4
    NOTIFICATION_ERROR_FINITE_STATE_MACHINE = 5
    NOTIFICATION_ERROR_CEASE = 6
    NOTIFICATION_ERROR_ROUTE_REFRESH_MESSAGE = 7

    # NOTIFICATION_ERROR_MESSAGE_HEADER sub-codes
    # According to RFC 4493 and RFC 4271
    # Unassigned: 4-255
    ##
    NOTIFICATION_ERROR_MESSAGE_HEADER_UNSPECIFIC = 0
    NOTIFICATION_ERROR_MESSAGE_HEADER_CONNECTION_NOT_SYNCHRONIZED = 1
    NOTIFICATION_ERROR_MESSAGE_HEADER_BAD_MESSAGE_LENGTH = 2
    NOTIFICATION_ERROR_MESSAGE_HEADER_BAD_MESSAGE_TYPE = 3

    # NOTIFICATION_ERROR_OPEN_MESSAGE sub-codes
    # According to RFC 4493, RFC 4271 and RFC 5492
    # Unassigned: 5 && 8-255
    ##
    NOTIFICATION_ERROR_OPEN_MESSAGE_UNSPECIFIC = 0
    NOTIFICATION_ERROR_OPEN_MESSAGE_UNSUPPORTED_VERSION_NUMBER = 1
    NOTIFICATION_ERROR_OPEN_MESSAGE_BAD_PEER_AS = 2
    NOTIFICATION_ERROR_OPEN_MESSAGE_BAD_BGP_IDENTIFIER = 3
    NOTIFICATION_ERROR_OPEN_MESSAGE_UNSUPPORTED_OPTIONAL_PARAMETER = 4
    NOTIFICATION_ERROR_OPEN_MESSAGE_UNACCEPTABLE_HOLD_TIME = 6
    NOTIFICATION_ERROR_OPEN_MESSAGE_UNSUPPORTED_CAPABILITY = 7

    # NOTIFICATION_ERROR_UPDATE_MESSAGE sub-codes
    # According to RFC 4493 and RFC 4271
    # Unassigned: 7 && 12-255
    ##
    NOTIFICATION_ERROR_UPDATE_MESSAGE_UNSPECIFIC = 0
    NOTIFICATION_ERROR_UPDATE_MESSAGE_MALFORMED_ATTRIBUTE_LIST = 1
    NOTIFICATION_ERROR_UPDATE_MESSAGE_UNRECOGNIZED_WELL_KNOWN_ATTRIBUTE = 2
    NOTIFICATION_ERROR_UPDATE_MESSAGE_MISSING_WELL_KNOWN_ATTRIBUTE = 3
    NOTIFICATION_ERROR_UPDATE_MESSAGE_ATTRIBUTE_FLAGS_ERROR = 4
    NOTIFICATION_ERROR_UPDATE_MESSAGE_ATTRIBUTE_LENGTH_ERROR = 5
    NOTIFICATION_ERROR_UPDATE_MESSAGE_INVALID_ORIGIN_ATTRIBUTE = 6
    NOTIFICATION_ERROR_UPDATE_MESSAGE_INVALID_NEXT_HOP_ATTRIBUTE = 8
    NOTIFICATION_ERROR_UPDATE_MESSAGE_OPTIONAL_ATTRIBUTE_ERROR = 9
    NOTIFICATION_ERROR_UPDATE_MESSAGE_INVALID_NETWORK_FIELD = 10
    NOTIFICATION_ERROR_UPDATE_MESSAGE_MALFORMED_AS_PATH = 11

    # NOTIFICATION_ERROR_FINITE_STATE_MACHINE sub-codes
    # According to RFC 6608
    # Unassigned: 4-255
    ##
    NOTIFICATION_ERROR_FINITE_STATE_MACHINE_UNSPECIFIC = 0
    NOTIFICATION_ERROR_FINITE_STATE_MACHINE_UNEXPECTED_MESSAGE_OPENSENT = 1
    NOTIFICATION_ERROR_FINITE_STATE_MACHINE_UNEXPECTED_MESSAGE_OPENCONFIRM = 2
    NOTIFICATION_ERROR_FINITE_STATE_MACHINE_UNEXPECTED_MESSAGE_ESTABLISHED = 3

    # NOTIFICATION_ERROR_CEASE sub-codes
    # According to RFC 4486
    # Unassigned: 9-255
    ##
    NOTIFICATION_ERROR_CEASE_RESERVED = 0
    NOTIFICATION_ERROR_CEASE_MAXIMUM_NUMBERS_OF_PREFIXES_REACHED = 1
    NOTIFICATION_ERROR_CEASE_ADMINISTRATIVE_SHUTDOWN = 2
    NOTIFICATION_ERROR_CEASE_PEER_DECONFIGURED = 3
    NOTIFICATION_ERROR_CEASE_ADMINISTRATIVE_RESET = 4
    NOTIFICATION_ERROR_CEASE_CONNECTION_REJECTED = 5
    NOTIFICATION_ERROR_CEASE_OTHER_CONFIGURATION_CHANGE = 6
    NOTIFICATION_ERROR_CEASE_CONNECTION_COLLISION_RESOLUTION = 7
    NOTIFICATION_ERROR_CEASE_OUT_OF_RESOURCES = 8

    # NOTIFICATION_ERROR_ROUTE_REFRESH_MESSAGE sub-codes
    # According to RFC 7313
    # Unassgined: 2-255
    ##
    NOTIFICATION_ERROR_ROUTE_REFRESH_MESSAGE_RESERVED = 0
    NOTIFICATION_ERROR_ROUTE_REFRESH_MESSAGE_INVALID_MESSAGE_LENGTH = 1

    # ** BGP Outbound Route Filtering (ORF) Types **
    # According to RF 5291
    # Unassigned: 1-63 && 66-127
    # Registration Procedurs:
    # :: 0-63       - Standards Action
    # :: 64-127     - First Come First Served
    # :: 128-255    - Vendor-Specific
    ##
    ORF_RESERVED = 0
    ORF_ADDRESS_PREFIX = 64
    ORF_CP = 65

    # ** BGP Tunnel Encapsulation Attribute Tunnel Types **
    # According to RFC 5512 and RFC 5566
    # Unassigned: 15-65535
    ##
    TUNNEL_ENCAPSULATION_ATTRIBUTE_RESERVED = 0
    TUNNEL_ENCAPSULATION_ATTRIBUTE_L2TPV3_OVER_IP = 1
    TUNNEL_ENCAPSULATION_ATTRIBUTE_GRE = 2
    TUNNEL_ENCAPSULATION_ATTRIBUTE_TRANSMIT_TUNNEL_ENDPOINT = 3
    TUNNEL_ENCAPSULATION_ATTRIBUTE_IPSEC_TUNNELMODE = 4
    TUNNEL_ENCAPSULATION_ATTRIBUTE_IP_IN_IP_TUNNEL_WITH_IPSEC_TRANSPORT_MODE = 5
    TUNNEL_ENCAPSULATION_ATTRIBUTE_MPLS_IN_IP_TUNNEL_WITH_IPSEC_TRANSPORT_MODE = 6
    TUNNEL_ENCAPSULATION_ATTRIBUTE_IP_IN_IP = 7
    TUNNEL_ENCAPSULATION_ATTRIBUTE_VXLAN = 8
    TUNNEL_ENCAPSULATION_ATTRIBUTE_NVGRE = 9
    TUNNEL_ENCAPSULATION_ATTRIBUTE_MPLS = 10
    TUNNEL_ENCAPSULATION_ATTRIBUTE_MPLS_IN_GRE = 11
    TUNNEL_ENCAPSULATION_ATTRIBUTE_VXLAN_GPE = 12
    TUNNEL_ENCAPSULATION_ATTRIBUTE_MPLS_UDP = 13
    TUNNEL_ENCAPSULATION_ATTRIBUTE_IPV6_TUNNEL = 14

    # ** TUNNEL_ENCAPSULATION_ATTRIBUTE sub-TLVS **
    # According to RFC 5512, RFC 5566 and RFC 5640
    # Unassigned: 6-255
    TUNNEL_ENCAPSULATION_ATTRIBUTE_TLV_RESERVED = 0
    TUNNEL_ENCAPSULATION_ATTRIBUTE_TLV_ENCAPSULATION = 1
    TUNNEL_ENCAPSULATION_ATTRIBUTE_TLV_PROTOCOL_TYPE = 2
    TUNNEL_ENCAPSULATION_ATTRIBUTE_TLV_IPSEC_TUNNEL_AUTH = 3
    TUNNEL_ENCAPSULATION_ATTRIBUTE_TLV_COLOR = 4
    TUNNEL_ENCAPSULATION_ATTRIBUTE_TLV_LOAD_BALANCING_BLOCK = 5

    # ** BGP Layer 2 Encapsulation Types **
    # According to RFC 6624, RFC 4446, RFC 4816, RFC 4618, RFC 4717,
    # RFC 3032, RFC 4619, RFC 4553, RFC 4761 and RFC 5086
    # Unassigned: 12-14 && 16 && 22-24 && 25-39 && 45-127 && 128-251
    # Registration procedures:
    # :: 0-127      - Expert Review
    # :: 128-251    - First Come First Served
    # :: 252-255    - Experimental Use
    ##
    LAYER2_ENCAPSULATION_RESERVED = 0
    LAYER2_ENCAPSULATION_FRAME_RELAY = 1
    LAYER2_ENCAPSULATION_ATM_AAL5_SDU_VCC_TRANSPORT = 2
    LAYER2_ENCAPSULATION_ATM_TRANSPARENT_CELL_TRANSPORT = 3
    LAYER2_ENCAPSULATION_ETHERNET_VLAN_TAGGED_MODE = 4
    LAYER2_ENCAPSULATION_ETHERNET_RAW_MODE = 5
    LAYER2_ENCAPSULATION_CISCO_HDLC = 6
    LAYER2_ENCAPSULATION_PPP = 7
    LAYER2_ENCAPSULATION_SONET_SDH_CIRCUIT_EMULATION_SERVICE = 8
    LAYER2_ENCAPSULATION_ATM_N_TO_ONE_VCC_CELL_TRANSPORT = 9
    LAYER2_ENCAPSULATION_ATM_N_TO_ONE_VPC_CELL_TRANSPORT = 10
    LAYER2_ENCAPSULATION_IP_LAYER2_TRANSPORT = 11
    LAYER2_ENCAPSULATION_FRAME_RELAY_PORT_MODE = 15
    LAYER2_ENCAPSULATION_STRUCTURE_AGNOSTIC_E1_OVER_PACKET = 17
    LAYER2_ENCAPSULATION_STRUCTURE_AGNOSTIC_T1_DS1_OVER_PACKET = 18
    LAYER2_ENCAPSULATION_VPLS = 19
    LAYER2_ENCAPSULATION_STRUCTURE_AGNOSTIC_T3_DS3_OVER_PACKET = 20
    LAYER2_ENCAPSULATION_NX64KBITS_BASIC_SERVICE_USING_STRUCTURE_AWARE = 21
    LAYER2_ENCAPSULATION_FRAME_RELAY_DLCI = 25
    LAYER2_ENCAPSULATION_STRUCTURE_AGNOSTIC_E3_OVER_PACKET = 40
    LAYER2_ENCAPSULATION_OCTET_ALIGNED_PAYLOAD_FOR_STRUCTURE_AGNOSTIC_DS1_CIRCUITS = 41
    LAYER2_ENCAPSULATION_E1_NX64KBITS_WITH_CAS_USING_STRUCTURE_AWARE = 42
    LAYER2_ENCAPSULATION_DS1_ESF_NX64KBITS_WITH_CAS_USING_STRUCTURE_AWARE = 43
    LAYER2_ENCAPSULATION_DS1_SF_NX64KBITS_WITH_CAS_USING_STRUCTURE_AWARE = 44

    # ** BGP Layer 2 TLV Types **
    # According to RFC 6624
    # Unassigned: 2-127 && 128-251
    # Registration procedures:
    # :: 0-127      - Expert Review
    # :: 128-251    - First Come First Served
    # :: 252-255    - Experimental Use
    ##
    LAYER2_TLV_RESERVED = 0
    LAYER2_TLV_CIRCUIT_STATUS_VECTOR = 1

    # ** BGP AIGP Attribute Types **
    # According to RFC 7311
    # Unassigned: 2-255
    ##
    AIGP_ATTRIBUTE_RESERVED = 0
    AIGP_ATTRIBUTE = 1

    # ** BGPKeepalive **
    ##
    KEEPALIVE_FIXED_LENGTH = 19 # (Every KEEPALIVE message must be 19 bytes long)

    # ** BGPRouteRefresh **
    # According to RFC 7313, RFC 2918 and RFC 5291
    # Unassigned: 3-254
    # Registration procedure:
    # :: 0-127      - Standards Action
    # :: 128-254    - First Come First Served
    ##
    ROUTE_REFRESH = 0
    ROUTE_REFRESH_BORR = 1
    ROUTE_REFRESH_EORR = 2

    # ** BGP P-Multicast Service Interface (PMSI) Tunnel Attribute Flags **
    # According to RFC 7902 and RFC 6514
    # Unassigned: 0 && 2
    # Temporary:
    # :: 3-4        - Assisted-Replication-Type (T) (Until: 2017-06-30)
    # :: 5          - Broadcast and Multicast (BM)  (Until: 2017-06-30)
    # :: 6          - Unknown (U)                   (Until: 2017-06-30)
    ##
    PMSI_TUNNEL_ATTRIBUTE_EXTENSION = 1
    PMSI_TUNNEL_ATTRIBUTE_LEAF_INFORMATION_REQUIRED = 7

    # ** BGPUpdate Flags **
    ##
    UPDATE_FLAG_LENGTH = 0x10
    UPDATE_FLAG_PARTIAL = 0x20
    UPDATE_FLAG_TRANSITIVE = 0x40
    UPDATE_FLAG_OPTIONAL = 0x80

    # ** BGPUpdate Path Attribute ORIGIN sub-codes **
    # According to RFC 4271
    ##
    ORIGIN_IGP = 0
    ORIGIN_EGP = 1
    ORIGIN_INCOMPLETE = 2

    # ** BGPUpdate Path Atribute AS_PATH sub-codes **
    # According to RFC 4271
    ##
    AS_PATH_SEGMENT_SET = 1
    AS_PATH_SEGMENT_SEQUENCE = 2

    # ** BGPUpdate Path Attribute EXTENDED_COMMUNITIES sub-types **
    # According to RFC 4360
    # NT = Non-Transitive
    # T = Transitive
    ##
    EXT_COMMUNITY_T_TWO_OCTET_AS_SPECIFIC = 0
    EXT_COMMUNITY_T_IPV4_ADDRESS_SPECIFIC = 1
    EXT_COMMUNITY_T_FOUR_OCTET_AS_SPECIFIC = 2
    EXT_COMMUNITY_T_OPAQUE = 3
    EXT_COMMUNITY_T_QOS_MARKING = 4
    EXT_COMMUNITY_T_COS_CAPABILITY = 5
    EXT_COMMUNITY_T_EVPN = 6
    EXT_COMMUNITY_T_FLOW_SPEC_REDIRECT_MIRROR_TO_IP_NEXT_HOP = 8
    EXT_COMMUNITY_T_GENERIC_EXPERIMENTAL_USE = 128
    EXT_COMMUNITY_T_GENERIC_EXPERIMENTAL_USE_PART2 = 129
    EXT_COMMUNITY_T_GENERIC_EXPERIMENTAL_USE_PART3 = 130

    EXT_COMMUNITY_NT_TWO_OCTET_AS_SPECIFIC = 64
    EXT_COMMUNITY_NT_IPV4_ADDRESS_SPECIFIC = 65
    EXT_COMMUNITY_NT_FOUR_OCTET_AS_SPECIFIC = 66
    EXT_COMMUNITY_NT_OPAQUE = 67
    EXT_COMMUNITY_NT_QOS_MARKING = 68

    # ** BGPUpdate Path Attribute EXTENDED_COMMUNITIES EVPN sub-types ***
    # According to RFC 7153
    ##
    EXT_COMMUNITY_EVPN_MAC_MOBILITY = 0
    EXT_COMMUNITY_EVPN_ESI_LABEL = 1
    EXT_COMMUNITY_EVPN_ES_IMPORT_ROUTE_TARGET = 2
    EXT_COMMUNITY_EVPN_ROUTERS_MAC = 3
    EXT_COMMUNITY_EVPN_LAYER2 = 4
    EXT_COMMUNITY_EVPN_ETREE = 5
    EXT_COMMUNITY_EVPN_DF_ELECTION = 6
    EXT_COMMUNITY_EVPN_ISID = 7

    # ** BGPUpdate Path Attribute EXTENDED_COMMUNITIES trans. Two-Octet AS-Specific **
    # According to RFC 4360, RFC 4577, RFC 4384, RFC 6514 and RFC 6074
    ##
    EXT_COMMUNITY_T_TWO_OCTET_AS_ROUTE_TARGET = 2
    EXT_COMMUNITY_T_TWO_OCTET_AS_ROUTE_ORIGIN = 3
    EXT_COMMUNITY_T_TWO_OCTET_AS_OSPF_DOMAIN_IDENTIFIER = 5
    EXT_COMMUNITY_T_TWO_OCTET_AS_BGP_DATA_COLLECTION = 8
    EXT_COMMUNITY_T_TWO_OCTET_AS_SOURCE_AS = 9
    EXT_COMMUNITY_T_TWO_OCTET_AS_L2VPN_IDENTIFIER = 10
    EXT_COMMUNITY_T_TWO_OCTET_AS_CISCO_VPN_DISTINGUISHER = 16
    EXT_COMMUNITY_T_TWO_OCTET_AS_ROUTE_TARGET_RECORD = 19

    # ** BGPUpdate Path Attribute EXTENDED_COMMUNITIES non-trans. Two-Octet AS-Specific **
    ##
    EXT_COMMUNITY_NT_TWO_OCTET_AS_LINK_BANDWIDTH = 4
    EXT_COMMUNITY_NT_TWO_OCTET_AS_VIRTUAL_NETWORK_IDENTIFIER = 128

    # ** BGPUpdate Path Attribute EXTENDED_COMMUNITIES trans. Four-Octet AS-Specific **
    # According to RFC 5668, RFC 4577, RFC 4384 and RFC 6514
    ##
    EXT_COMMUNITY_T_FOUR_OCTET_AS_ROUTE_TARGET = 2
    EXT_COMMUNITY_T_FOUR_OCTET_AS_ROUTE_ORIGIN = 3
    EXT_COMMUNITY_T_FOUR_OCTET_AS_GENERIC = 4
    EXT_COMMUNITY_T_FOUR_OCTET_AS_OSPF_DOMAIN_IDENTIFIER = 5
    EXT_COMMUNITY_T_FOUR_OCTET_AS_BGP_DATA_COLLECTION = 8
    EXT_COMMUNITY_T_FOUR_OCTET_AS_SOURCE_AS = 9
    EXT_COMMUNITY_T_FOUR_OCTET_AS_CISCO_VPN_IDENTIFIER = 16
    EXT_COMMUNITY_T_FOUR_OCTET_AS_ROUTE_TARGET_RECORD = 19

    # ** BGPUpdate Path Attribute EXTENDED_COMMUNITIES non-trans. Four-Octet AS-Specific **
    ##
    EXT_COMMUNITY_NT_FOUR_OCTET_AS_GENERIC = 4

    # ** BGPUpdate Path Attribute EXTENDED_COMMUNITIES trans. IPv4-Address-Specific **
    # According to RFC 4360, RFC 4577, RFC 6074, RFC 6541 and RFC 7524
    ##
    EXT_COMMUNITY_T_IPV4_ADDRESS_ROUTE_TARGET = 2
    EXT_COMMUNITY_T_IPV4_ADDRESS_ROUTE_ORIGIN = 3
    EXT_COMMUNITY_T_IPV4_ADDRESS_OSPF_DOMAIN_IDENTIFIER = 5
    EXT_COMMUNITY_T_IPV4_ADDRESS_OSPF_ROUTE_ID = 7
    EXT_COMMUNITY_T_IPV4_ADDRESS_L2VPN_IDENTIFIER = 10
    EXT_COMMUNITY_T_IPV4_ADDRESS_VRF_ROUTE_IMPORT = 11
    EXT_COMMUNITY_T_IPV4_ADDRESS_FLOW_SPEC_REDIRECT_TO_IPV4 = 12
    EXT_COMMUNITY_T_IPV4_ADDRESS_CISCO_VPN_DISTINGUISHER = 16
    EXT_COMMUNITY_T_IPV4_ADDRESS_INTER_AREA_P2MP_SEGMENTED_NEXT_HOP = 18
    EXT_COMMUNITY_T_IPV4_ADDRESS_ROUTE_TARGET_RECORD = 19

    # ** BGPUpdate Path Attribute EXTENDED_COMMUNITIES trans. Opaque **
    ##
    EXT_COMMUNITY_T_OPAQUE_COST_COMMUNITY = 1
    EXT_COMMUNITY_T_OPAQUE_CP_ORF = 3
    EXT_COMMUNITY_T_OPAQUE_EXTRANET_SOURCE = 4
    EXT_COMMUNITY_T_OPAQUE_EXTRANET_SEPARATION = 5
    EXT_COMMUNITY_T_OPAQUE_OSPF_ROUTE_TYPE = 6
    EXT_COMMUNITY_T_OPAQUE_ADDITIONAL_PMSI_TUNNEL_ATTRIBUTE_FLAGS = 7
    EXT_COMMUNITY_T_OPAQUE_COLOR_EXTENDED = 11
    EXT_COMMUNITY_T_OPAQUE_ENCAPSULATION = 12
    EXT_COMMUNITY_T_OPAQUE_DEFAULT_GATEWAY = 13
    EXT_COMMUNITY_T_OPAQUE_PPMP_LABEL = 14
    EXT_COMMUNITY_T_OPAQUE_CONSISTENT_HASH_SORT_ORDER = 20

    # ** BGPUpdate Path Attribute EXTENDED_COMMUNITIES non-trans. Opaque **
    ##
    EXT_COMMUNITY_NT_OPAQUE_BGP_ORIGIN_VALIDATION_STATE = 0
    EXT_COMMUNITY_NT_OPAQUE_COST_COMMUNITY = 1

    # ** BGPUpdate Path Attribute EXTENDED_COMMUNITIES trans. Generic Experimental **
    ##
    EXT_COMMUNITY_T_GENERIC_EXPERIMENTAL_OSPF_ROUTE_TYPE = 0
    EXT_COMMUNITY_T_GENERIC_EXPERIMENTAL_OSPF_ROUTER_ID = 1
    EXT_COMMUNITY_T_GENERIC_EXPERIMENTAL_OSPF_DOMAIN_IDENTIFIER = 5
    EXT_COMMUNITY_T_GENERIC_EXPERIMENTAL_FLOW_SPEC_TRAFFIC_RATE = 6
    EXT_COMMUNITY_T_GENERIC_EXPERIMENTAL_FLOW_SPEC_TRAFFIC_ACTION = 7
    EXT_COMMUNITY_T_GENERIC_EXPERIMENTAL_FLOW_SEPC_REDIRECT_AS_2BYTE_FORMAT = 8
    EXT_COMMUNITY_T_GENERIC_EXPERIMENTAL_FLOW_SPEC_TRAFFIC_REMARKING = 9
    EXT_COMMUNITY_T_GENERIC_EXPERIMENTAL_LAYER2_INFO = 10
    EXT_COMMUNITY_T_GENERIC_EXPERIMENTAL_E_TREE_INFO = 11

    # ** BGPUpdate Path Attribute EXTENDED_COMMUNITIES trans. Generic Experimental Part 2 **
    ##
    EXT_COMMUNITY_T_GENERIC_EXPERIMENTAL_PART2_FLOW_SPEC_REDIRECT_IPV4_FORMAT = 8

    # ** BGPUpdate Path Attribute EXTENDED_COMMUNITIES trans. Generic Experimental Part 3 **
    EXT_COMMUNITY_T_GENERIC_EXPERIMENTAL_PART3_FLOW_SPEC_AS_4BYTE_FORMAT = 8



#
# This file is part of PCAP BGP Parser (pbgpp)
#
# Copyright 2016-2017 DE-CIX Management GmbH
# Author: Tobias Hannaske <tobias.hannaske@de-cix.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from pbgpp.BGP.Update.Route import BGPRoute
from pbgpp.Output.Filter import BGPFilter
from pbgpp.BGP.Statics import BGPStatics


class NextHopFilter(BGPFilter):
    def __init__(self, values=[]):
        BGPFilter.__init__(self, values)

    def apply(self, message):
        try:
            # NEXT_HOP is a path attribute of BGP UPDATE message
            # Therefore we first need to make that we are currently handling an UPDATE message
            if message.type is not BGPStatics.MESSAGE_TYPE_UPDATE:
                # Skip messages that are no UPDATE messages
                return None

            for attribute in message.path_attributes:
                # Skip attributes that are no NEXT_HOP attributes
                if attribute.type is not BGPStatics.UPDATE_ATTRIBUTE_NEXT_HOP:
                    continue

                # Here we found the NEXT_HOP attribute - loop through allowed values
                for value in self.values:
                    negated = False
                    if value[0:1] == "~":
                        negated = True
                        value = value[1:]

                    if not negated and str(BGPRoute.decimal_ip_to_string(attribute.next_hop)) == str(value):
                        # Match on NEXT_HOP attribute - Return message
                        return message

                    if negated and str(BGPRoute.decimal_ip_to_string(attribute.next_hop)) != str(value):
                        return message

            # Searched value was not found
            return None
        except Exception as e:
            # On error the filtering was not successful (due to wrong fields, etc.)
            return None



#
# This file is part of PCAP BGP Parser (pbgpp)
#
# Copyright 2016-2017 DE-CIX Management GmbH
# Author: Tobias Hannaske <tobias.hannaske@de-cix.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from pbgpp.Output.Filter import BGPFilter


class TimestampFilter(BGPFilter):
    def __init__(self, values=[]):
        BGPFilter.__init__(self, values)

    def apply(self, pcap_information):
        # !!! Attention: This is a pre-parsing filter!
        # This filter must be applied BEFORE parsing, otherwise it will unnecessarily slow down
        # the whole application. BGP messages don't have to be parsed when applying that filter
        # directly after reading PCAP packet header

        try:
            for v in self.values:
                negated = False
                if v[0:1] == "~":
                    negated = True
                    v = v[1:]

                if "." in v:
                    if not negated and str(pcap_information.get_timestamp()[0]) + "." + str(pcap_information.get_timestamp()[1]) == v:
                        return True

                    if negated and str(pcap_information.get_timestamp()[0]) + "." + str(pcap_information.get_timestamp()[1]) != v:
                        return True
                else:
                    if not negated and int(pcap_information.get_timestamp()[0]) == int(v):
                        return True

                    if negated and int(pcap_information.get_timestamp()[0]) != int(v):
                        return True

            # Searched value was not found
            return False
        except Exception as e:
            # On error the filtering was not successful (due to wrong fields, etc.)
            return False



#
# This file is part of PCAP BGP Parser (pbgpp)
#
# Copyright 2016-2017 DE-CIX Management GmbH
# Author: Tobias Hannaske <tobias.hannaske@de-cix.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


class BGPFormatter:
    # Define formatter types
    FORMATTER_HUMAN_READABLE = 0
    FORMATTER_JSON = 1
    FORMATTER_LINE_BASED = 2

    # Define available formatters here
    CHOICES_FORMATTERS = ["HUMAN_READABLE", "JSON", "LINE_BASED"]



#
# This file is part of PCAP BGP Parser (pbgpp)
#
# Copyright 2016-2017 DE-CIX Management GmbH
# Author: Tobias Hannaske <tobias.hannaske@de-cix.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from pbgpp.BGP.Exceptions import BGPUpdateASPathSegmentFactoryError
from pbgpp.BGP.Statics import BGPStatics
from pbgpp.BGP.Translation import BGPTranslation


class BGPUpdateASPathSegment:
    def __init__(self, segment_type, segments):
        self.segment_type = segment_type
        self.segments = segments

    @classmethod
    def factory(cls, segment_type, segments):
        if segment_type == BGPStatics.AS_PATH_SEGMENT_SET:
            return cls(segment_type, segments)

        if segment_type == BGPStatics.AS_PATH_SEGMENT_SEQUENCE:
            return cls(segment_type, segments)

        raise BGPUpdateASPathSegmentFactoryError("could not validate segment type.")

    def json(self):
        r = {
            "segment_type": self.segment_type,
            "segment_type_string": BGPTranslation.path_segment_type(self.segment_type),
            "segments": []
        }

        for s in self.segments:
            r["segments"].append(s)

        return r

    def __str__(self):
        # Display AS_SEQUENCE in brackets (it's an ordered list of ASN)
        # Display AS_SET as raw numbers
        return_string = ""

        if self.segment_type == BGPStatics.AS_PATH_SEGMENT_SEQUENCE:
            first_asn = True

            for asn in self.segments:
                if first_asn:
                    return_string += str(asn)
                    first_asn = False
                    continue

                return_string += " " + str(asn)

        elif self.segment_type == BGPStatics.AS_PATH_SEGMENT_SET:
            # Brackets because AS_SEQUENCE is an ordered list
            return_string += "("

            first_asn = True

            for asn in self.segments:
                if first_asn:
                    return_string += str(asn)
                    first_asn = False
                    continue

                return_string += " " + str(asn)

            # Close brackets
            return_string += ")"

        else:
            pass

        return None if len(return_string) == 0 else return_string



#
# This file is part of PCAP BGP Parser (pbgpp)
#
# Copyright 2016-2017 DE-CIX Management GmbH
# Author: Tobias Hannaske <tobias.hannaske@de-cix.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from pbgpp.BGP.Message import BGPMessage
from pbgpp.Output.Exceptions import OutputHandlerError
from pbgpp.Output.Formatter import BGPFormatter
from pbgpp.Output.Pipe import BGPPipe


class OutputHandler:

    def __init__(self, message, filter=[], formatter=None, pipe=None):
        # Pre-check variables
        if not isinstance(message, BGPMessage):
            raise OutputHandlerError("packet must be instance of BGPPacket.")

        if not isinstance(formatter, BGPFormatter):
            raise OutputHandlerError("formatter must be instance of Formatters.")

        if not isinstance(pipe, BGPPipe):
            raise OutputHandlerError("pipe must be instance of Pipe.")

        # Assign class variables
        self.message = message

        self.filter = filter
        self.formatter = formatter
        self.pipe = pipe

        self.output = None

    def __filter(self):
        # Apply runtime filter - maybe the message isn't even required to be displayed.
        # In this case we don't need to waste computing time on formatting and piping.
        # Filters are always connected with a logical AND. One filter is able to allow multiple values
        # for one specific filter. Those values are linked with a logical OR.
        # Example: --filter-next-hop 11.11.11.11 --filter-next-hop 12.12.12.12 --filter-source-ip 13.13.13.13
        # Example: Those filters will filter packets that have
        # Example: ((next-hop == 11.11.11.11)
        # Example: _OR_ (next-hop == 12.12.12.12))
        # Example: _AND_ (source-ip == 13.13.13.13)

        for f in self.filter:
            if self.message is None:
                # We don't need to try to apply filters on a message that was already set to None
                break

            # Apply post-parsing filter to parsed packet
            self.message = f.apply(self.message)

    def __format(self):
        # Format the packet or a single message (e.g. apply JSON-formatting)
        try:
            self.output = self.formatter.apply(self.message)
        except TypeError:
            self.output = None
        except AttributeError:
            self.output = None

    def __pipe(self):
        # Pipe the filtered and formatted output (e.g. into a file or into stdout)
        self.pipe.output(self.output)

    def handle(self):
        # Filters will set self.packet to None if no filter will apply
        if self.message is not None:
            self.__filter()

            if self.message is not None:
                self.__format()

                if self.output is not None:
                    self.__pipe()
                else:
                    # Don't display message if there occurred an error during formatting
                    pass

            else:
                # Don't display message if there occurred an error during filtering
                # or if the packet is set to None during filtering
                pass



#
# This file is part of PCAP BGP Parser (pbgpp)
#
# Copyright 2016-2017 DE-CIX Management GmbH
# Author: Tobias Hannaske <tobias.hannaske@de-cix.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import struct


class BGPUpdateLargeCommunity:
    def __init__(self, global_administrator, local_data_part_1, local_data_part_2):
        self.global_administrator = global_administrator
        self.local_data_part_1 = local_data_part_1
        self.local_data_part_2 = local_data_part_2

    def _get_four_octect_values_as_str(self):
        if isinstance(self.global_administrator, int) and isinstance(self.local_data_part_1, int) and isinstance(self.local_data_part_2, int):
            return str(self.global_administrator), str(self.local_data_part_1), str(self.local_data_part_2)
        elif isinstance(self.global_administrator, bytes) and isinstance(self.local_data_part_1, bytes) and isinstance(self.local_data_part_2, bytes):
            global_administrator = str(struct.unpack("!L", self.global_administrator)[0])
            local_data_part_1 = str(struct.unpack("!L", self.local_data_part_1)[0])
            local_data_part_2 = str(struct.unpack("!L", self.local_data_part_2)[0])
            return global_administrator, local_data_part_1, local_data_part_2
        else:
            return None

    def __str__(self):
        four_octect_values = self._get_four_octect_values_as_str()

        if four_octect_values:
            return "{}:{}:{}".format(
                four_octect_values[0],
                four_octect_values[1],
                four_octect_values[2]
            )
        else:
            return None

    def json(self):
        four_octect_values = self._get_four_octect_values_as_str()

        if four_octect_values:
            return {
                "global_administrator": four_octect_values[0],
                "local_data_part_1": four_octect_values[1],
                "local_data_part_2": four_octect_values[2]
            }
        else:
            return None



#
# This file is part of PCAP BGP Parser (pbgpp)
#
# Copyright 2016-2017 DE-CIX Management GmbH
# Author: Tobias Hannaske <tobias.hannaske@de-cix.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import struct

from pbgpp.BGP.Statics import BGPStatics
from pbgpp.BGP.Translation import BGPTranslation
from pbgpp.BGP.Update.ASPathSegment import BGPUpdateASPathSegment
from pbgpp.BGP.Update.PathAttribute import BGPPathAttribute


class PathAttributeASPath(BGPPathAttribute):
    def __init__(self, payload):
        BGPPathAttribute.__init__(self, payload)
        self.type = BGPStatics.UPDATE_ATTRIBUTE_AS_PATH

        # Path attribute specific variables
        self.asn_byte_length = None
        self.path_segments = []

        self.__parse()

    def __str__(self):
        # Display AS_SEQUENCE in brackets (it's an ordered list of ASN)
        # Display AS_SET as raw numbers
        return_string = ""
        first = True

        for segment in self.path_segments:
            if first:
                first = False
                return_string += str(segment)
                continue

            return_string += " " + str(segment)

        return "" if len(return_string) == 0 else return_string

    def __parse(self):
        try:
            self.parsed = True
            self.error = False

            if len(self.payload) > 0:
                self.asn_byte_length = PathAttributeASPath.as_heuristic(self.payload)

                # Handle payload if ASNs are 4 bytes long
                if self.asn_byte_length == 4:
                    continue_loop = True
                    current_byte_position = 0

                    while continue_loop:
                        fields = struct.unpack("!BB", self.payload[current_byte_position:current_byte_position + 2])
                        segment_type = fields[0]
                        segment_length = fields[1]

                        segment_fields = struct.unpack("!" + ("I" * segment_length), self.payload[current_byte_position + 2:current_byte_position + 2 + (segment_length * 4)])

                        self.path_segments.append(BGPUpdateASPathSegment.factory(segment_type, segment_fields))

                        current_byte_position += 2 + (segment_length * 4)
                        if current_byte_position >= len(self.payload):
                            continue_loop = False

                # Handle payload if ASNs are 2 bytes long
                elif self.asn_byte_length == 2:
                    continue_loop = True
                    current_byte_position = 0

                    while continue_loop:
                        fields = struct.unpack("!BB", self.payload[current_byte_position:current_byte_position + 2])
                        segment_type = fields[0]
                        segment_length = fields[1]

                        segment_fields = struct.unpack("!" + ("H" * segment_length), self.payload[current_byte_position + 2:current_byte_position + 2 + (segment_length * 2)])

                        self.path_segments.append(BGPUpdateASPathSegment.factory(segment_type, segment_fields))

                        current_byte_position += 2 + (segment_length * 2)
                        if current_byte_position >= len(self.payload):
                            continue_loop = False

                else:
                    # Could not determine the correct byte length of ASN
                    # This SHOULD never happen, but there is no safe way to determine the length
                    self.error = True

            else:
                # There is nothing to parse due to empty payload
                pass

        except Exception as e:
            self.error = True

    def json(self):
        r = {
            "asn_byte_length": self.asn_byte_length,
            "type": self.type,
            "type_string": BGPTranslation.path_attribute(self.type),
            "path_segments": [],
            "error": self.error,
        }

        for segment in self.path_segments:
            r["path_segments"].append(segment.json())

        return r

    @staticmethod
    def as_heuristic(payload):
        # This function is kind of complicated
        # Problem: There is no clear way to determine if an ASN has two or four bytes length
        # This code is translated from C into Python from Wireshark project
        # @todo Documentation of ASN heuristic code
        current_byte_position = 1

        length = struct.unpack("!B", payload[current_byte_position:current_byte_position + 1])[0]
        current_byte_position += 1

        offset_check = current_byte_position + (2 * length)

        try:
            next_type = struct.unpack("!B", payload[offset_check:offset_check + 1])[0]
        except Exception as e:
            # There is no next type
            next_type = -1

        if offset_check == len(payload):
            assumed_as_length = 2
        elif (next_type == 1) or (next_type == 2) or (next_type == 3) or (next_type == 4):
            j = 0
            asn_is_null = 0
            while (j < length) and (asn_is_null == 0):
                try:
                    check_value = \
                    struct.unpack("!H", payload[current_byte_position + (2 * j):(current_byte_position + (2 * j) + 2)])[0]
                    if check_value == 0:
                        asn_is_null = 1
                except:
                    break
                finally:
                    j += 1

            if asn_is_null == 0:
                assumed_as_length = 2
            else:
                assumed_as_length = 4
        else:
            assumed_as_length = 4

        k = 0
        while k < len(payload):
            k += 1
            length = struct.unpack("!B", payload[k:k + 1])[0]
            k += 1
            k += length * assumed_as_length

        if k == len(payload):
            return assumed_as_length
        else:
            return -1



#
# This file is part of PCAP BGP Parser (pbgpp)
#
# Copyright 2016-2018 DE-CIX Management GmbH
# Author: Benedikt Rudolph <benedikt.rudolph@de-cix.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from pbgpp.BGP.Update.Route import BGPRoute
from pbgpp.Output.Filter import BGPFilter
from pbgpp.BGP.Statics import BGPStatics

# This filter combines the two alternative ways to signal a blackhole routing
# for a particular  prefix.
#   * The legacy method is to set the next_hop attribute to a special IP address.
#   * The preferred method is to set a well-known BGP Community value (RFC 7999).
# The code for this filter is a combination of CommunityValueFilter and NextHopFilter.
class BlackholeFilter(BGPFilter):
    def __init__(self, values=[]):
        BGPFilter.__init__(self, values)

    def apply(self, message):
        try:
            # NEXT_HOP and COMMUNITIES are attributes of a BGP UPDATE message
            if message.type is not BGPStatics.MESSAGE_TYPE_UPDATE:
                # Skip messages that are no UPDATE messages
                return None

            for attribute in message.path_attributes:
                # Skip attributes that are no NEXT_HOP attributes
                if attribute.type is BGPStatics.UPDATE_ATTRIBUTE_NEXT_HOP:
                    # Here we found the NEXT_HOP attribute - check for blackhole next_hop
                    for value in self.values:
                        if str(BGPRoute.decimal_ip_to_string(attribute.next_hop)) == str(value):
                            # Match on NEXT_HOP attribute - Return message
                            return message

                # Alternatively check if well-known BGP community is set (RFC 7999)
                if attribute.type is BGPStatics.UPDATE_ATTRIBUTE_COMMUNITIES:

                    for community in attribute.communities:
                        if community.asn == 65535 and community.value == 666:
                          return message

            # Searched value was not found
            return None
        except Exception as e:
            # On error the filtering was not successful (due to wrong fields, etc.)
            return None



#
# This file is part of PCAP BGP Parser (pbgpp)
#
# Copyright 2016-2017 DE-CIX Management GmbH
# Author: Tobias Hannaske <tobias.hannaske@de-cix.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import struct
from pbgpp.PCAP.Information import PCAPLayer3Information


class PCAPIP:
    PROTO_TCP = 0x0006
    BITMASK_IP_HEADER_LENGTH = 0xf

    def __init__(self, payload):
        # Assign variables
        self.payload = payload
        self.protocol = None
        self.addresses = None

        self.header_length = None
        self.version = None
        self.total_length = None

        # Start parsing
        self.__parse()

    def __parse(self):
        version_length = struct.unpack("!B", self.payload[:1])[0]

        # IP header length and version is packed into 1 byte (8 bit)
        self.header_length = (version_length & self.BITMASK_IP_HEADER_LENGTH) * 4
        self.version = (version_length >> 4)

        self.total_length = struct.unpack("!H", self.payload[2:4])[0]
        self.protocol = struct.unpack("!B", self.payload[9:10])[0]

        ip_set = struct.unpack("!BBBBBBBB", self.payload[12:20])
        self.addresses = PCAPLayer3Information(ip_set[0:4], ip_set[4:8])

    def get_protocol(self):
        return self.protocol

    def get_addresses(self):
        return self.addresses

    def get_ip_payload(self):
        return self.payload[self.header_length:self.total_length]



#
# This file is part of PCAP BGP Parser (pbgpp)
#
# Copyright 2016-2017 DE-CIX Management GmbH
# Author: Tobias Hannaske <tobias.hannaske@de-cix.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import socket
import struct

from pbgpp.BGP.Exceptions import BGPRouteInitializeError, BGPRouteConvertionError


class BGPRoute:
    def __init__(self, prefix, prefix_length):
        # A route is universally used
        # Prefix = e.g. 123.123.123.123
        # Length = 32
        # To String: 123.123.123.123/32 (CIDR notation)

        # Assign values
        self.prefix = prefix
        self.prefix_length = prefix_length

        # Values that need to be assigned due to parsing
        self.prefix_string = None
        self.prefix_length_string = None
        self.prefix_length_decimal = None

        self._parse()

    @classmethod
    def from_binary(cls, prefix, prefix_length):
        # Create a class instance from bytes
        if isinstance(prefix, bytes) and isinstance(prefix_length, bytes):
            return cls(prefix, prefix_length)
        else:
            raise BGPRouteInitializeError("prefix and prefix_length must be instance of bytes.")

    def __str__(self):
        # Return the prefix string that was created during parsing
        return self.prefix_string

    def __eq__(self, other):
        # Compare two routes by comparing the prefix and its length
        if isinstance(other, BGPRoute):
            if self.prefix == other.prefix and self.prefix_length == other.prefix_length:
                return True
        else:
            # This wont work for any other classes. Just for BGPRoute objects.
            return NotImplemented

    def _parse(self):
        # Check the prefix length at first as that length is needed to determine
        # how many bytes we need to parse afterwards
        self.prefix_length_decimal = struct.unpack("!B", self.prefix_length)[0]
        self.prefix_length_string = str(self.prefix_length_decimal)

        if 0 <= self.prefix_length_decimal <= 8:
            # Length of prefix field: 1 Byte
            fields = struct.unpack("!B", self.prefix)
            self.prefix_string = str(fields[0]) + ".0.0.0/" + self.prefix_length_string

        elif 9 <= self.prefix_length_decimal <= 16:
            # Length of prefix field: 2 Bytes
            fields = struct.unpack("!BB", self.prefix)
            self.prefix_string = str(fields[0]) + "." + str(fields[1]) + ".0.0/" + self.prefix_length_string

        elif 17 <= self.prefix_length_decimal <= 24:
            # Length of prefix field: 3 Bytes
            fields = struct.unpack("!BBB", self.prefix)
            self.prefix_string = str(fields[0]) + "." + str(fields[1]) + "." + str(fields[2]) + ".0/" + self.prefix_length_string

        elif 25 <= self.prefix_length_decimal:
            # Length of prefix field: 4 Bytes
            fields = struct.unpack("!BBBB", self.prefix)
            self.prefix_string = str(fields[0]) + "." + str(fields[1]) + "." + str(fields[2]) + "." + str(fields[3]) + "/" + self.prefix_length_string

        else:
            raise BGPRouteConvertionError("was not able to parse bytes.")

    @staticmethod
    def decimal_ip_to_string(decimal):
        return socket.inet_ntoa(struct.pack('!L', decimal))
